"""
Execution module for creating shortcuts on Windows. Handles file shortcuts
(`.lnk`) and url shortcuts (`.url`). Allows for the configuration of icons and
hot keys on file shortcuts. Changing the icon and hot keys are unsupported for
url shortcuts.

.. versionadded:: 3005
"""

# https://docs.microsoft.com/en-us/troubleshoot/windows-client/admin-development/create-desktop-shortcut-with-wsh
# https://docs.microsoft.com/en-us/previous-versions/windows/internet-explorer/ie-developer/windows-scripting/f5y78918(v=vs.84)
import logging
import os
import time

import salt.utils.path
import salt.utils.platform
import salt.utils.winapi
from salt.exceptions import CommandExecutionError

HAS_WIN32 = False
if salt.utils.platform.is_windows():
    import win32com.client

    HAS_WIN32 = True

# Set up logging
log = logging.getLogger(__name__)

# Define the module's virtual name
__virtualname__ = "shortcut"


WINDOW_STYLE = {
    1: "Normal",
    3: "Maximized",
    7: "Minimized",
    "Normal": 1,
    "Maximized": 3,
    "Minimized": 7,
}


def __virtual__():
    """
    Make sure we're on Windows
    """
    # Verify Windows
    if not salt.utils.platform.is_windows():
        log.debug("Shortcut module only available on Windows systems")
        return False, "Shortcut module only available on Windows systems"
    if not HAS_WIN32:
        log.debug("Shortcut module requires pywin32")
        return False, "Shortcut module requires pywin32"

    return __virtualname__


def get(path):
    r"""
    Gets the properties for a shortcut

    Args:

        path (str):
            The path to the shortcut. Must have a `.lnk` or `.url` file
            extension.

    Returns:
        dict: A dictionary containing all available properties for the specified
            shortcut

    CLI Example:

    .. code-block:: bash

        salt '*' shortcut.get path='C:\path\to\shortcut.lnk'
    """
    if not os.path.exists(path):
        raise CommandExecutionError(f"Shortcut not found: {path}")

    if not path.endswith((".lnk", ".url")):
        _, ext = os.path.splitext(path)
        raise CommandExecutionError(f"Invalid file extension: {ext}")

    # This will load the existing shortcut
    with salt.utils.winapi.Com():
        shell = win32com.client.Dispatch(  # pylint: disable=used-before-assignment
            "WScript.Shell"
        )
        shortcut = shell.CreateShortcut(path)

        arguments = ""
        description = ""
        hot_key = ""
        icon_location = ""
        icon_index = 0
        window_style = ""
        working_dir = ""

        path = salt.utils.path.expand(shortcut.FullName)
        # A shortcut can have either a .lnk or a .url extension. We only want to
        # expand the target if it is a .lnk
        if path.endswith(".lnk"):
            target = shortcut.TargetPath
            if target:
                target = salt.utils.path.expand(target)
            else:
                msg = f"Not a valid shortcut: {path}"
                log.debug(msg)
                raise CommandExecutionError(msg)
            if shortcut.Arguments:
                arguments = shortcut.Arguments
            if shortcut.Description:
                description = shortcut.Description
            if shortcut.Hotkey:
                hot_key = shortcut.Hotkey
            if shortcut.IconLocation:
                icon_location, icon_index = shortcut.IconLocation.split(",")
                if icon_location:
                    icon_location = salt.utils.path.expand(icon_location)
            if shortcut.WindowStyle:
                window_style = WINDOW_STYLE[shortcut.WindowStyle]
            if shortcut.WorkingDirectory:
                working_dir = salt.utils.path.expand(shortcut.WorkingDirectory)
        else:
            target = shortcut.TargetPath

        return {
            "arguments": arguments,
            "description": description,
            "hot_key": hot_key,
            "icon_index": int(icon_index),
            "icon_location": icon_location,
            "path": path,
            "target": target,
            "window_style": window_style,
            "working_dir": working_dir,
        }


def _set_info(
    path,
    target="",
    arguments="",
    description="",
    hot_key="",
    icon_index=0,
    icon_location="",
    window_style="Normal",
    working_dir="",
):
    r"""
    The main worker function for creating and modifying shortcuts. the `create`
    and `modify` functions are wrappers around this function.

    Args:

        path (str): The full path to the shortcut.

        target (:obj:`str`, optional): The full path to the target.

            Defaults is "".

        arguments (:obj:`str`, optional):
            Any arguments to be passed to the target.

            Default is "".

        description (:obj:`str`, optional):
            The description for the shortcut. This is shown in the ``Comment``
            field of the dialog box.

            Default is "".

        hot_key (:obj:`str`, optional):
            A combination of hot Keys to trigger this shortcut. This is
            something like ``Ctrl+Alt+D``. This is shown in the ``Shortcut key``
            field in the dialog box. Available options are:

            - Ctrl
            - Alt
            - Shift
            - Ext

            Default is "".

        icon_index (:obj:`int`, optional):
            The index for the icon to use in files that contain multiple icons.

            Default is 0.

        icon_location (:obj`str`, optional):
            The full path to a file containing icons. This is shown in the
            ``Change Icon`` dialog box by clicking the ``Change Icon`` button.
            If no file is specified and a binary is passed as the target,
            Windows will attempt to get the icon from the binary file.

            Default is "".

        window_style (:obj:`str`, optional):
            The window style the program should start in. This is shown in the
            ``Run`` field of the dialog box. Valid options are:

            - Normal
            - Minimized
            - Maximized

            Default is "Normal".

        working_dir (:obj:`str`, optional):
            The full path to the working directory for the program to run in.
            This is shown in the ``Start in`` field of the dialog box.

            Default is "".

    Returns:
        bool: ``True`` if successful.
    """
    path = salt.utils.path.expand(path)

    # This will load the existing shortcut if it already exists
    # If it is a new shortcut, it won't be created until it is saved
    with salt.utils.winapi.Com():
        shell = win32com.client.Dispatch("WScript.Shell")
        shortcut = shell.CreateShortcut(path)

        # A shortcut can have either a .lnk or a .url extension. We only want to
        # expand the target if it is a .lnk
        if path.endswith(".lnk"):
            if target:
                target = salt.utils.path.expand(target)

            # These settings only apply to lnk shortcuts
            if arguments:
                shortcut.Arguments = arguments
            if description:
                shortcut.Description = description
            if hot_key:
                shortcut.Hotkey = hot_key
            if icon_location:
                shortcut.IconLocation = ",".join([icon_location, str(icon_index)])
            if window_style:
                shortcut.WindowStyle = WINDOW_STYLE[window_style]
            if working_dir:
                shortcut.WorkingDirectory = working_dir

        shortcut.TargetPath = target

        shortcut.Save()
    return True


def modify(
    path,
    target="",
    arguments="",
    description="",
    hot_key="",
    icon_index=0,
    icon_location="",
    window_style="Normal",
    working_dir="",
):
    r"""
    Modify an existing shortcut. This can be a file shortcut (``.lnk``) or a
    url shortcut (``.url``).

    Args:

        path (str): The full path to the shortcut. Must have a `.lnk` or `.url`
            file extension.

        target (:obj:`str`, optional): The full path to the target.

            Default is "".

        arguments (:obj:`str`, optional):
            Any arguments to be passed to the target.

            Default is "".

        description (:obj:`str`, optional):
            The description for the shortcut. This is shown in the ``Comment``
            field of the dialog box.

            Default is "".

        hot_key (:obj:`str`, optional):
            A combination of hot Keys to trigger this shortcut. This is
            something like ``Ctrl+Alt+D``. This is shown in the ``Shortcut key``
            field in the dialog box. Available options are:

            - Ctrl
            - Alt
            - Shift
            - Ext

            Default is "".

        icon_index (:obj:`int`, optional):
            The index for the icon to use in files that contain multiple icons.

            Default is 0.

        icon_location (:obj:`str`, optional):
            The full path to a file containing icons. This is shown in the
            ``Change Icon`` dialog box by clicking the ``Change Icon`` button.
            If no file is specified and a binary is passed as the target,
            Windows will attempt to get the icon from the binary file.

            Default is "".

        window_style (:obj:`str`, optional):
            The window style the program should start in. This is shown in the
            ``Run`` field of the dialog box. Valid options are:

            - Normal
            - Minimized
            - Maximized

            Default is "Normal".

        working_dir (:obj:`str`, optional):
            The full path to the working directory for the program to run in.
            This is shown in the ``Start in`` field of the dialog box.

            Default is "".

    Returns:
        bool: ``True`` if successful.

    CLI Example:

    .. code-block:: bash

        # Modify an existing shortcut. Set it to target notepad.exe
        salt '*' shortcut.modify 'C:\path\to\shortcut.lnk' 'C:\Windows\notepad.exe'
    """
    if not os.path.exists(path):
        raise CommandExecutionError(f"Shortcut not found: {path}")

    if not path.endswith((".lnk", ".url")):
        _, ext = os.path.splitext(path)
        raise CommandExecutionError(f"Invalid file extension: {ext}")

    return _set_info(
        path=path,
        arguments=arguments,
        description=description,
        hot_key=hot_key,
        icon_index=icon_index,
        icon_location=icon_location,
        target=target,
        window_style=window_style,
        working_dir=working_dir,
    )


def create(
    path,
    target,
    arguments="",
    description="",
    hot_key="",
    icon_index=0,
    icon_location="",
    window_style="Normal",
    working_dir="",
    backup=False,
    force=False,
    make_dirs=False,
    user=None,
):
    r"""
    Create a new shortcut. This can be a file shortcut (``.lnk``) or a url
    shortcut (``.url``).

    Args:

        path (str):
            The full path to the shortcut. Must have a `.lnk` or `.url` file
            extension.

        target (str): The full path to the target.

        arguments (:obj:`str`, optional):
            Any arguments to be passed to the target.

            Default is "".

        description (:obj:`str`, optional):
            The description for the shortcut. This is shown in the ``Comment``
            field of the dialog box.

            Default is "".

        hot_key (:obj:`str`, optional):
            A combination of hot Keys to trigger this shortcut. This is
            something like ``Ctrl+Alt+D``. This is shown in the ``Shortcut key``
            field in the dialog box. Available options are:

            - Ctrl
            - Alt
            - Shift
            - Ext

            Default is "".

        icon_index (:obj:`int`, optional):
            The index for the icon to use in files that contain multiple icons.

            Default is 0.

        icon_location (:obj:`str`, optional):
            The full path to a file containing icons. This is shown in the
            ``Change Icon`` dialog box by clicking the ``Change Icon`` button.
            If no file is specified and a binary is passed as the target,
            Windows will attempt to get the icon from the binary file.

            Default is "".

        window_style (:obj:`str`, optional):
            The window style the program should start in. This is shown in the
            ``Run`` field of the dialog box. Valid options are:

            - Normal
            - Minimized
            - Maximized

            Default is "Normal".

        working_dir (:obj:`str`, optional):
            The full path to the working directory for the program to run in.
            This is shown in the ``Start in`` field of the dialog box.

            Default is "".

        backup (:obj:`bool`, optional):
            If there is already a shortcut with the same name, set this value to
            ``True`` to backup the existing shortcut and continue creating the
            new shortcut.

            Default is ``False``.

        force (:obj:`bool`, optional):
            If there is already a shortcut with the same name and you aren't
            backing up the shortcut, set this value to ``True`` to remove the
            existing shortcut and create a new with these settings.

            Default is ``False``.

        make_dirs (:obj:`bool`, optional):
            If the parent directory structure does not exist for the new
            shortcut, create it.

            Default is ``False``.

        user (:obj:`str`, optional):
            The user to be the owner of any directories created by setting
            ``make_dirs`` to ``True``. If no value is passed, Salt will use the
            user account that it is running under.

            Default is "".

    Returns:
        bool: ``True`` if successful.

    Raises:
        CommandExecutionError: If the path is not a ``.lnk`` or ``.url`` file
            extension.
        CommandExecutionError: If there is an existing shortcut with the same
            name and ``backup`` and ``force`` are both ``False``
        CommandExecutionError: If the parent directory is not created and
            ``make_dirs`` is ``False``
        CommandExecutionError: If there was an error creating the parent
            directories

    CLI Example:

    .. code-block:: bash

        # Create a shortcut and set the ``Shortcut key`` (``hot_key``)
        salt '*' shortcut.create 'C:\path\to\shortcut.lnk' 'C:\Windows\notepad.exe' hot_key='Ctrl+Alt+N'

        # Create a shortcut and change the icon to the 3rd one in the icon file
        salt '*' shortcut.create 'C:\path\to\shortcut.lnk' 'C:\Windows\notepad.exe' icon_location='C:\path\to\icon.ico' icon_index=2

        # Create a shortcut and change the startup mode to full screen
        salt '*' shortcut.create 'C:\path\to\shortcut.lnk' 'C:\Windows\notepad.exe' window_style='Maximized'

        # Create a shortcut and change the icon
        salt '*' shortcut.create 'C:\path\to\shortcut.lnk' 'C:\Windows\notepad.exe' icon_location='C:\path\to\icon.ico'

        # Create a shortcut and force it to overwrite an existing shortcut
        salt '*' shortcut.create 'C:\path\to\shortcut.lnk' 'C:\Windows\notepad.exe' force=True

        # Create a shortcut and create any parent directories if they are missing
        salt '*' shortcut.create 'C:\path\to\shortcut.lnk' 'C:\Windows\notepad.exe' make_dirs=True
    """
    if not path.endswith((".lnk", ".url")):
        _, ext = os.path.splitext(path)
        raise CommandExecutionError(f"Invalid file extension: {ext}")

    if os.path.exists(path):
        if backup:
            log.debug("Backing up: %s", path)
            file, ext = os.path.splitext(path)
            ext = ext.strip(".")
            backup_path = f"{file}-{time.time_ns()}.{ext}"
            os.rename(path, backup_path)
        elif force:
            log.debug("Removing: %s", path)
            os.remove(path)
        else:
            log.debug("Shortcut exists: %s", path)
            raise CommandExecutionError("Found existing shortcut")

    if not os.path.isdir(os.path.dirname(path)):
        if make_dirs:

            # Get user from opts if not defined
            if not user:
                user = __opts__["user"]

            # Make sure the user exists in Windows
            # Salt default is 'SYSTEM' for Windows
            if not __salt__["user.info"](user):
                # User not found, use the account salt is running under
                # If username not found, use System
                user = __salt__["user.current"]()
                if not user:
                    user = "SYSTEM"

            try:
                __salt__["file.makedirs"](path=path, owner=user)
            except CommandExecutionError as exc:
                raise CommandExecutionError(
                    f"Error creating parent directory: {exc.message}"
                )
        else:
            raise CommandExecutionError(
                f"Parent directory not present: {os.path.dirname(path)}"
            )

    return _set_info(
        path=path,
        arguments=arguments,
        description=description,
        hot_key=hot_key,
        icon_index=icon_index,
        icon_location=icon_location,
        target=target,
        window_style=window_style,
        working_dir=working_dir,
    )



import importlib
import logging
import re
from textwrap import dedent

import pytest

import salt.client.ssh.client
import salt.client.ssh.shell as shell
import salt.config
import salt.roster
import salt.utils.files
import salt.utils.path
import salt.utils.platform
import salt.utils.thin
import salt.utils.yaml
from salt.client import ssh
from tests.support.mock import MagicMock, call, patch

log = logging.getLogger(__name__)


@pytest.fixture
def opts(master_opts):
    master_opts["argv"] = [
        "ssh.set_auth_key",
        "root",
        "hobn+amNAXSBTiOXEqlBjGB...rsa root@master",
    ]
    return master_opts


@pytest.fixture()
def mock_bin_paths():
    with patch("salt.utils.path.which") as mock_which:
        mock_which.side_effect = lambda x: {
            "ssh-keygen": "ssh-keygen",
            "ssh": "ssh",
            "scp": "scp",
        }.get(x, None)
        importlib.reload(shell)
        yield
    importlib.reload(shell)


@pytest.fixture
def target():
    return {
        "passwd": "abc123",
        "ssh_options": None,
        "sudo": False,
        "identities_only": False,
        "host": "login1",
        "user": "root",
        "timeout": 65,
        "remote_port_forwards": None,
        "sudo_user": "",
        "port": "22",
        "priv": "/etc/salt/pki/master/ssh/salt-ssh.rsa",
    }


def test_single_opts(opts, target, mock_bin_paths):
    """Sanity check for ssh.Single options"""

    single = ssh.Single(
        opts,
        opts["argv"],
        "localhost",
        mods={},
        fsclient=None,
        thin=salt.utils.thin.thin_path(opts["cachedir"]),
        mine=False,
        **target,
    )

    assert single.shell._ssh_opts() == ""
    expected_cmd = (
        "ssh login1 "
        "-o KbdInteractiveAuthentication=no -o "
        "PasswordAuthentication=yes -o ConnectTimeout=65 -o Port=22 "
        "-o IdentityFile=/etc/salt/pki/master/ssh/salt-ssh.rsa "
        "-o User=root  date +%s"
    )
    assert single.shell._cmd_str("date +%s") == expected_cmd


def test_run_with_pre_flight(opts, target, tmp_path):
    """
    test Single.run() when ssh_pre_flight is set
    and script successfully runs
    """
    target["ssh_pre_flight"] = str(tmp_path / "script.sh")
    single = ssh.Single(
        opts,
        opts["argv"],
        "localhost",
        mods={},
        fsclient=None,
        thin=salt.utils.thin.thin_path(opts["cachedir"]),
        mine=False,
        **target,
    )

    cmd_ret = ("Success", "", 0)
    mock_flight = MagicMock(return_value=cmd_ret)
    mock_cmd = MagicMock(return_value=cmd_ret)
    patch_flight = patch("salt.client.ssh.Single.run_ssh_pre_flight", mock_flight)
    patch_cmd = patch("salt.client.ssh.Single.cmd_block", mock_cmd)
    patch_exec_cmd = patch(
        "salt.client.ssh.shell.Shell.exec_cmd", return_value=("", "", 1)
    )
    patch_os = patch("os.path.exists", side_effect=[True])

    with patch_os, patch_flight, patch_cmd, patch_exec_cmd:
        ret = single.run()
        mock_cmd.assert_called()
        mock_flight.assert_called()
        assert ret == cmd_ret


def test_run_with_pre_flight_with_args(opts, target, tmp_path):
    """
    test Single.run() when ssh_pre_flight is set
    and script successfully runs
    """
    target["ssh_pre_flight"] = str(tmp_path / "script.sh")
    target["ssh_pre_flight_args"] = "foobar"
    single = ssh.Single(
        opts,
        opts["argv"],
        "localhost",
        mods={},
        fsclient=None,
        thin=salt.utils.thin.thin_path(opts["cachedir"]),
        mine=False,
        **target,
    )

    cmd_ret = ("Success", "foobar", 0)
    mock_flight = MagicMock(return_value=cmd_ret)
    mock_cmd = MagicMock(return_value=cmd_ret)
    patch_flight = patch("salt.client.ssh.Single.run_ssh_pre_flight", mock_flight)
    patch_cmd = patch("salt.client.ssh.Single.cmd_block", mock_cmd)
    patch_exec_cmd = patch(
        "salt.client.ssh.shell.Shell.exec_cmd", return_value=("", "", 1)
    )
    patch_os = patch("os.path.exists", side_effect=[True])

    with patch_os, patch_flight, patch_cmd, patch_exec_cmd:
        ret = single.run()
        mock_cmd.assert_called()
        mock_flight.assert_called()
        assert ret == cmd_ret


def test_run_with_pre_flight_stderr(opts, target, tmp_path):
    """
    test Single.run() when ssh_pre_flight is set
    and script errors when run
    """
    target["ssh_pre_flight"] = str(tmp_path / "script.sh")
    single = ssh.Single(
        opts,
        opts["argv"],
        "localhost",
        mods={},
        fsclient=None,
        thin=salt.utils.thin.thin_path(opts["cachedir"]),
        mine=False,
        **target,
    )

    cmd_ret = ("", "Error running script", 1)
    mock_flight = MagicMock(return_value=cmd_ret)
    mock_cmd = MagicMock(return_value=cmd_ret)
    patch_flight = patch("salt.client.ssh.Single.run_ssh_pre_flight", mock_flight)
    patch_cmd = patch("salt.client.ssh.Single.cmd_block", mock_cmd)
    patch_exec_cmd = patch(
        "salt.client.ssh.shell.Shell.exec_cmd", return_value=("", "", 1)
    )
    patch_os = patch("os.path.exists", side_effect=[True])

    with patch_os, patch_flight, patch_cmd, patch_exec_cmd:
        ret = single.run()
        mock_cmd.assert_not_called()
        mock_flight.assert_called()
        assert ret == cmd_ret


def test_run_with_pre_flight_script_doesnot_exist(opts, target, tmp_path):
    """
    test Single.run() when ssh_pre_flight is set
    and the script does not exist
    """
    target["ssh_pre_flight"] = str(tmp_path / "script.sh")
    single = ssh.Single(
        opts,
        opts["argv"],
        "localhost",
        mods={},
        fsclient=None,
        thin=salt.utils.thin.thin_path(opts["cachedir"]),
        mine=False,
        **target,
    )

    cmd_ret = ("Success", "", 0)
    mock_flight = MagicMock(return_value=cmd_ret)
    mock_cmd = MagicMock(return_value=cmd_ret)
    patch_flight = patch("salt.client.ssh.Single.run_ssh_pre_flight", mock_flight)
    patch_cmd = patch("salt.client.ssh.Single.cmd_block", mock_cmd)
    patch_exec_cmd = patch(
        "salt.client.ssh.shell.Shell.exec_cmd", return_value=("", "", 1)
    )
    patch_os = patch("os.path.exists", side_effect=[False])

    with patch_os, patch_flight, patch_cmd, patch_exec_cmd:
        ret = single.run()
        mock_cmd.assert_called()
        mock_flight.assert_not_called()
        assert ret == cmd_ret


def test_run_with_pre_flight_thin_dir_exists(opts, target, tmp_path):
    """
    test Single.run() when ssh_pre_flight is set
    and thin_dir already exists
    """
    target["ssh_pre_flight"] = str(tmp_path / "script.sh")
    single = ssh.Single(
        opts,
        opts["argv"],
        "localhost",
        mods={},
        fsclient=None,
        thin=salt.utils.thin.thin_path(opts["cachedir"]),
        mine=False,
        **target,
    )

    cmd_ret = ("", "", 0)
    mock_flight = MagicMock(return_value=cmd_ret)
    mock_cmd = MagicMock(return_value=cmd_ret)
    patch_flight = patch("salt.client.ssh.Single.run_ssh_pre_flight", mock_flight)
    patch_cmd = patch("salt.client.ssh.shell.Shell.exec_cmd", mock_cmd)
    patch_cmd_block = patch("salt.client.ssh.Single.cmd_block", mock_cmd)
    patch_os = patch("os.path.exists", return_value=True)

    with patch_os, patch_flight, patch_cmd, patch_cmd_block:
        ret = single.run()
        mock_cmd.assert_called()
        mock_flight.assert_not_called()
        assert ret == cmd_ret


def test_run_ssh_pre_flight(opts, target, tmp_path):
    """
    test Single.run_ssh_pre_flight function
    """
    target["ssh_pre_flight"] = str(tmp_path / "script.sh")
    single = ssh.Single(
        opts,
        opts["argv"],
        "localhost",
        mods={},
        fsclient=None,
        thin=salt.utils.thin.thin_path(opts["cachedir"]),
        mine=False,
        **target,
    )

    cmd_ret = ("Success", "", 0)
    mock_flight = MagicMock(return_value=cmd_ret)
    mock_cmd = MagicMock(return_value=cmd_ret)
    patch_flight = patch("salt.client.ssh.Single.run_ssh_pre_flight", mock_flight)
    patch_cmd = patch("salt.client.ssh.Single.cmd_block", mock_cmd)
    patch_exec_cmd = patch(
        "salt.client.ssh.shell.Shell.exec_cmd", return_value=("", "", 1)
    )
    patch_os = patch("os.path.exists", side_effect=[True])

    with patch_os, patch_flight, patch_cmd, patch_exec_cmd:
        ret = single.run()
        mock_cmd.assert_called()
        mock_flight.assert_called()
        assert ret == cmd_ret


def test_execute_script(opts, target, tmp_path):
    """
    test Single.execute_script()
    """
    single = ssh.Single(
        opts,
        opts["argv"],
        "localhost",
        mods={},
        fsclient=None,
        thin=salt.utils.thin.thin_path(opts["cachedir"]),
        mine=False,
        winrm=False,
        **target,
    )

    exp_ret = ("Success", "", 0)
    mock_cmd = MagicMock(return_value=exp_ret)
    patch_cmd = patch("salt.client.ssh.shell.Shell.exec_cmd", mock_cmd)
    script = str(tmp_path / "script.sh")

    with patch_cmd:
        ret = single.execute_script(script=script)
        assert ret == exp_ret
        assert mock_cmd.call_count == 2
        assert [
            call(f"/bin/sh '{script}'"),
            call(f"rm '{script}'"),
        ] == mock_cmd.call_args_list


def test_shim_cmd(opts, target, tmp_path):
    """
    test Single.shim_cmd()
    """
    single = ssh.Single(
        opts,
        opts["argv"],
        "localhost",
        mods={},
        fsclient=None,
        thin=salt.utils.thin.thin_path(opts["cachedir"]),
        mine=False,
        winrm=False,
        tty=True,
        **target,
    )

    exp_ret = ("Success", "", 0)
    mock_cmd = MagicMock(return_value=exp_ret)
    patch_cmd = patch("salt.client.ssh.shell.Shell.exec_cmd", mock_cmd)
    patch_send = patch("salt.client.ssh.shell.Shell.send", return_value=("", "", 0))
    patch_rand = patch("os.urandom", return_value=b"5\xd9l\xca\xc2\xff")
    tmp_file = tmp_path / "tmp_file"
    mock_tmp = MagicMock()
    patch_tmp = patch("tempfile.NamedTemporaryFile", mock_tmp)
    mock_tmp.return_value.__enter__.return_value.name = tmp_file

    with patch_cmd, patch_tmp, patch_send:
        ret = single.shim_cmd(cmd_str="echo test")
        assert ret == exp_ret
        assert [
            call(f"/bin/sh '.{tmp_file.name}'"),
            call(f"rm '.{tmp_file.name}'"),
        ] == mock_cmd.call_args_list


def test_shim_cmd_copy_fails(opts, target, caplog):
    """
    test Single.shim_cmd() when copying the file fails
    """
    single = ssh.Single(
        opts,
        opts["argv"],
        "localhost",
        mods={},
        fsclient=None,
        thin=salt.utils.thin.thin_path(opts["cachedir"]),
        mine=False,
        winrm=False,
        tty=True,
        **target,
    )

    ret_cmd = ("Success", "", 0)
    mock_cmd = MagicMock(return_value=ret_cmd)
    patch_cmd = patch("salt.client.ssh.shell.Shell.exec_cmd", mock_cmd)
    ret_send = ("", "General error in file copy", 1)
    patch_send = patch("salt.client.ssh.shell.Shell.send", return_value=ret_send)
    patch_rand = patch("os.urandom", return_value=b"5\xd9l\xca\xc2\xff")

    with patch_cmd, patch_rand, patch_send:
        ret = single.shim_cmd(cmd_str="echo test")
        assert ret == ret_send
        assert "Could not copy the shim script to target" in caplog.text
        mock_cmd.assert_not_called()


def test_run_ssh_pre_flight_no_connect(opts, target, tmp_path, caplog, mock_bin_paths):
    """
    test Single.run_ssh_pre_flight when you
    cannot connect to the target
    """
    pre_flight = tmp_path / "script.sh"
    pre_flight.write_text("")
    target["ssh_pre_flight"] = str(pre_flight)
    single = ssh.Single(
        opts,
        opts["argv"],
        "localhost",
        mods={},
        fsclient=None,
        thin=salt.utils.thin.thin_path(opts["cachedir"]),
        mine=False,
        winrm=False,
        tty=True,
        **target,
    )
    mock_exec_cmd = MagicMock(return_value=("", "", 1))
    patch_exec_cmd = patch("salt.client.ssh.shell.Shell.exec_cmd", mock_exec_cmd)
    tmp_file = tmp_path / "tmp_file"
    mock_tmp = MagicMock()
    patch_tmp = patch("tempfile.NamedTemporaryFile", mock_tmp)
    mock_tmp.return_value.__enter__.return_value.name = tmp_file
    ret_send = (
        "",
        "ssh: connect to host 192.168.1.186 port 22: No route to host\nscp: Connection closed\n",
        255,
    )
    send_mock = MagicMock(return_value=ret_send)
    patch_send = patch("salt.client.ssh.shell.Shell.send", send_mock)

    with caplog.at_level(logging.TRACE):
        with patch_send, patch_exec_cmd, patch_tmp:
            ret = single.run_ssh_pre_flight()

    # Flush the logging handler just to be sure
    caplog.handler.flush()

    assert "Copying the pre flight script" in caplog.text
    assert "Could not copy the pre flight script to target" in caplog.text
    assert ret == ret_send
    assert send_mock.call_args_list[0][0][0] == tmp_file
    target_script = send_mock.call_args_list[0][0][1]
    assert re.search(r".[a-z0-9]+", target_script)
    mock_exec_cmd.assert_not_called()


def test_run_ssh_pre_flight_permission_denied(opts, target, tmp_path):
    """
    test Single.run_ssh_pre_flight when you
    cannot copy script to the target due to
    a permission denied error
    """
    pre_flight = tmp_path / "script.sh"
    pre_flight.write_text("")
    target["ssh_pre_flight"] = str(pre_flight)
    single = ssh.Single(
        opts,
        opts["argv"],
        "localhost",
        mods={},
        fsclient=None,
        thin=salt.utils.thin.thin_path(opts["cachedir"]),
        mine=False,
        winrm=False,
        tty=True,
        **target,
    )
    mock_exec_cmd = MagicMock(return_value=("", "", 1))
    patch_exec_cmd = patch("salt.client.ssh.shell.Shell.exec_cmd", mock_exec_cmd)
    tmp_file = tmp_path / "tmp_file"
    mock_tmp = MagicMock()
    patch_tmp = patch("tempfile.NamedTemporaryFile", mock_tmp)
    mock_tmp.return_value.__enter__.return_value.name = tmp_file
    ret_send = (
        "",
        'scp: dest open "/tmp/preflight.sh": Permission denied\nscp: failed to upload file /etc/salt/preflight.sh to /tmp/preflight.sh\n',
        255,
    )
    send_mock = MagicMock(return_value=ret_send)
    patch_send = patch("salt.client.ssh.shell.Shell.send", send_mock)

    with patch_send, patch_exec_cmd, patch_tmp:
        ret = single.run_ssh_pre_flight()
    assert ret == ret_send
    assert send_mock.call_args_list[0][0][0] == tmp_file
    target_script = send_mock.call_args_list[0][0][1]
    assert re.search(r".[a-z0-9]+", target_script)
    mock_exec_cmd.assert_not_called()


def test_run_ssh_pre_flight_connect(opts, target, tmp_path, caplog, mock_bin_paths):
    """
    test Single.run_ssh_pre_flight when you
    can connect to the target
    """
    pre_flight = tmp_path / "script.sh"
    pre_flight.write_text("")
    target["ssh_pre_flight"] = str(pre_flight)
    single = ssh.Single(
        opts,
        opts["argv"],
        "localhost",
        mods={},
        fsclient=None,
        thin=salt.utils.thin.thin_path(opts["cachedir"]),
        mine=False,
        winrm=False,
        tty=True,
        **target,
    )
    ret_exec_cmd = ("", "", 1)
    mock_exec_cmd = MagicMock(return_value=ret_exec_cmd)
    patch_exec_cmd = patch("salt.client.ssh.shell.Shell.exec_cmd", mock_exec_cmd)
    tmp_file = tmp_path / "tmp_file"
    mock_tmp = MagicMock()
    patch_tmp = patch("tempfile.NamedTemporaryFile", mock_tmp)
    mock_tmp.return_value.__enter__.return_value.name = tmp_file
    ret_send = (
        "",
        "\rroot@192.168.1.187's password: \n\rpreflight.sh 0%    0 0.0KB/s   --:-- ETA\rpreflight.sh 100%   20     2.7KB/s   00:00 \n",
        0,
    )
    send_mock = MagicMock(return_value=ret_send)
    patch_send = patch("salt.client.ssh.shell.Shell.send", send_mock)

    with caplog.at_level(logging.TRACE):
        with patch_send, patch_exec_cmd, patch_tmp:
            ret = single.run_ssh_pre_flight()

    # Flush the logging handler just to be sure
    caplog.handler.flush()

    assert "Executing the pre flight script on target" in caplog.text
    assert ret == ret_exec_cmd
    assert send_mock.call_args_list[0][0][0] == tmp_file
    target_script = send_mock.call_args_list[0][0][1]
    assert re.search(r".[a-z0-9]+", target_script)
    mock_exec_cmd.assert_called()


def test_run_ssh_pre_flight_shutil_fails(opts, target, tmp_path):
    """
    test Single.run_ssh_pre_flight when cannot
    copyfile with shutil
    """
    pre_flight = tmp_path / "script.sh"
    pre_flight.write_text("")
    target["ssh_pre_flight"] = str(pre_flight)
    single = ssh.Single(
        opts,
        opts["argv"],
        "localhost",
        mods={},
        fsclient=None,
        thin=salt.utils.thin.thin_path(opts["cachedir"]),
        mine=False,
        winrm=False,
        tty=True,
        **target,
    )
    ret_exec_cmd = ("", "", 1)
    mock_exec_cmd = MagicMock(return_value=ret_exec_cmd)
    patch_exec_cmd = patch("salt.client.ssh.shell.Shell.exec_cmd", mock_exec_cmd)
    tmp_file = tmp_path / "tmp_file"
    mock_tmp = MagicMock()
    patch_tmp = patch("tempfile.NamedTemporaryFile", mock_tmp)
    mock_tmp.return_value.__enter__.return_value.name = tmp_file
    send_mock = MagicMock()
    mock_shutil = MagicMock(side_effect=IOError("Permission Denied"))
    patch_shutil = patch("shutil.copyfile", mock_shutil)
    patch_send = patch("salt.client.ssh.shell.Shell.send", send_mock)

    with patch_send, patch_exec_cmd, patch_tmp, patch_shutil:
        ret = single.run_ssh_pre_flight()

    assert ret == (
        "",
        "Could not copy pre flight script to temporary path",
        1,
    )
    mock_exec_cmd.assert_not_called()
    send_mock.assert_not_called()


@pytest.mark.skip_on_windows(reason="SSH_PY_SHIM not set on windows")
@pytest.mark.slow_test
def test_cmd_run_set_path(opts, target):
    """
    test when set_path is set
    """
    target["set_path"] = "$PATH:/tmp/path/"
    single = ssh.Single(
        opts,
        opts["argv"],
        "localhost",
        mods={},
        fsclient=None,
        thin=salt.utils.thin.thin_path(opts["cachedir"]),
        mine=False,
        **target,
    )

    ret = single._cmd_str()
    assert re.search("\\" + target["set_path"], ret)


@pytest.mark.skip_on_windows(reason="SSH_PY_SHIM not set on windows")
@pytest.mark.slow_test
def test_cmd_run_not_set_path(opts, target):
    """
    test when set_path is not set
    """
    single = ssh.Single(
        opts,
        opts["argv"],
        "localhost",
        mods={},
        fsclient=None,
        thin=salt.utils.thin.thin_path(opts["cachedir"]),
        mine=False,
        **target,
    )

    ret = single._cmd_str()
    assert re.search('SET_PATH=""', ret)


@pytest.mark.skip_on_windows(reason="SSH_PY_SHIM not set on windows")
@pytest.mark.slow_test
def test_cmd_block_python_version_error(opts, target):
    single = ssh.Single(
        opts,
        opts["argv"],
        "localhost",
        mods={},
        fsclient=None,
        thin=salt.utils.thin.thin_path(opts["cachedir"]),
        mine=False,
        winrm=False,
        **target,
    )
    mock_shim = MagicMock(
        return_value=(("", "ERROR: Unable to locate appropriate python command\n", 10))
    )
    patch_shim = patch("salt.client.ssh.Single.shim_cmd", mock_shim)
    with patch_shim:
        ret = single.cmd_block()
        assert "ERROR: Python version error. Recommendation(s) follow:" in ret[0]


def _check_skip(grains):
    if grains["os"] == "MacOS":
        return True
    return False


@pytest.mark.skip_initial_gh_actions_failure(skip=_check_skip)
@pytest.mark.skip_on_windows(reason="pre_flight_args is not implemented for Windows")
@pytest.mark.parametrize(
    "test_opts",
    [
        (None, ""),
        ("one", " one"),
        ("one two", " one two"),
        ("| touch /tmp/test", " '|' touch /tmp/test"),
        ("; touch /tmp/test", " ';' touch /tmp/test"),
        (["one"], " one"),
        (["one", "two"], " one two"),
        (["one", "two", "| touch /tmp/test"], " one two '| touch /tmp/test'"),
        (["one", "two", "; touch /tmp/test"], " one two '; touch /tmp/test'"),
    ],
)
def test_run_with_pre_flight_args(opts, target, test_opts, tmp_path):
    """
    test Single.run() when ssh_pre_flight is set
    and script successfully runs
    """
    opts["ssh_run_pre_flight"] = True
    pre_flight_script = tmp_path / "script.sh"
    pre_flight_script.write_text("")
    target["ssh_pre_flight"] = str(pre_flight_script)

    if test_opts[0] is not None:
        target["ssh_pre_flight_args"] = test_opts[0]
    expected_args = test_opts[1]

    single = ssh.Single(
        opts,
        opts["argv"],
        "localhost",
        mods={},
        fsclient=None,
        thin=salt.utils.thin.thin_path(opts["cachedir"]),
        mine=False,
        **target,
    )

    cmd_ret = ("Success", "", 0)
    mock_cmd = MagicMock(return_value=cmd_ret)
    mock_exec_cmd = MagicMock(return_value=("", "", 0))
    patch_cmd = patch("salt.client.ssh.Single.cmd_block", mock_cmd)
    patch_exec_cmd = patch("salt.client.ssh.shell.Shell.exec_cmd", mock_exec_cmd)
    patch_shell_send = patch(
        "salt.client.ssh.shell.Shell.send", return_value=("", "", 0)
    )
    patch_os = patch("os.path.exists", side_effect=[True])

    with patch_os, patch_cmd, patch_exec_cmd, patch_shell_send:
        single.run()
        script_args = mock_exec_cmd.mock_calls[0].args[0]
        assert re.search(r"\/bin\/sh '.[a-z0-9]+", script_args)


@pytest.mark.slow_test
@pytest.mark.skip_on_windows(reason="Windows does not support salt-ssh")
@pytest.mark.skip_if_binaries_missing("ssh", check_all=True)
def test_ssh_single__cmd_str(opts):
    argv = []
    id_ = "minion"
    host = "minion"

    single = ssh.Single(opts, argv, id_, host, sudo=False)
    cmd = single._cmd_str()
    expected = dedent(
        """
        SUDO=""
        if [ -n "" ]
        then SUDO=" "
        fi
        SUDO_USER=""
        if [ "$SUDO" ] && [ "$SUDO_USER" ]
        then SUDO="$SUDO -u $SUDO_USER"
        fi
        """
    )

    assert expected in cmd


@pytest.mark.slow_test
@pytest.mark.skip_on_windows(reason="Windows does not support salt-ssh")
@pytest.mark.skip_if_binaries_missing("ssh", check_all=True)
def test_ssh_single__cmd_str_sudo(opts):
    argv = []
    id_ = "minion"
    host = "minion"

    single = ssh.Single(opts, argv, id_, host, sudo=True)
    cmd = single._cmd_str()
    expected = dedent(
        """
        SUDO=""
        if [ -n "sudo" ]
        then SUDO="sudo "
        fi
        SUDO_USER=""
        if [ "$SUDO" ] && [ "$SUDO_USER" ]
        then SUDO="$SUDO -u $SUDO_USER"
        fi
        """
    )

    assert expected in cmd


@pytest.mark.slow_test
@pytest.mark.skip_on_windows(reason="Windows does not support salt-ssh")
@pytest.mark.skip_if_binaries_missing("ssh", check_all=True)
def test_ssh_single__cmd_str_sudo_user(opts):
    argv = []
    id_ = "minion"
    host = "minion"
    user = "wayne"

    single = ssh.Single(opts, argv, id_, host, sudo=True, sudo_user=user)
    cmd = single._cmd_str()
    expected = dedent(
        """
        SUDO=""
        if [ -n "sudo" ]
        then SUDO="sudo "
        fi
        SUDO_USER="wayne"
        if [ "$SUDO" ] && [ "$SUDO_USER" ]
        then SUDO="$SUDO -u $SUDO_USER"
        fi
        """
    )

    assert expected in cmd


@pytest.mark.slow_test
@pytest.mark.skip_on_windows(reason="Windows does not support salt-ssh")
@pytest.mark.skip_if_binaries_missing("ssh", check_all=True)
def test_ssh_single__cmd_str_sudo_passwd(opts):
    argv = []
    id_ = "minion"
    host = "minion"
    passwd = "salty"

    single = ssh.Single(opts, argv, id_, host, sudo=True, passwd=passwd)
    cmd = single._cmd_str()
    expected = dedent(
        """
        SUDO=""
        if [ -n "sudo -p '[salt:sudo:d11bd4221135c33324a6bdc09674146fbfdf519989847491e34a689369bbce23]passwd:'" ]
        then SUDO="sudo -p '[salt:sudo:d11bd4221135c33324a6bdc09674146fbfdf519989847491e34a689369bbce23]passwd:' "
        fi
        SUDO_USER=""
        if [ "$SUDO" ] && [ "$SUDO_USER" ]
        then SUDO="$SUDO -u $SUDO_USER"
        fi
        """
    )

    assert expected in cmd


@pytest.mark.slow_test
@pytest.mark.skip_on_windows(reason="Windows does not support salt-ssh")
@pytest.mark.skip_if_binaries_missing("ssh", check_all=True)
def test_ssh_single__cmd_str_sudo_passwd_user(opts):
    argv = []
    id_ = "minion"
    host = "minion"
    user = "wayne"
    passwd = "salty"

    single = ssh.Single(opts, argv, id_, host, sudo=True, passwd=passwd, sudo_user=user)
    cmd = single._cmd_str()
    expected = dedent(
        """
        SUDO=""
        if [ -n "sudo -p '[salt:sudo:d11bd4221135c33324a6bdc09674146fbfdf519989847491e34a689369bbce23]passwd:'" ]
        then SUDO="sudo -p '[salt:sudo:d11bd4221135c33324a6bdc09674146fbfdf519989847491e34a689369bbce23]passwd:' "
        fi
        SUDO_USER="wayne"
        if [ "$SUDO" ] && [ "$SUDO_USER" ]
        then SUDO="$SUDO -u $SUDO_USER"
        fi
        """
    )

    assert expected in cmd



import pytest

import salt.utils.platform

from . import normalize_ret

pytestmark = [
    pytest.mark.windows_whitelisted,
    pytest.mark.core_test,
]


def test_watch_in(state, state_tree):
    """
    test watch_in requisite when there is a success
    """
    sls_contents = """
    return_changes:
      test.succeed_with_changes:
        - watch_in:
          - test: watch_states

    watch_states:
      test.succeed_without_changes
    """
    changes = "test_|-return_changes_|-return_changes_|-succeed_with_changes"
    watch = "test_|-watch_states_|-watch_states_|-succeed_without_changes"
    with pytest.helpers.temp_file("requisite.sls", sls_contents, state_tree):
        ret = state.sls("requisite")
        assert ret[changes].full_return["__run_num__"] == 0
        assert ret[changes].changes["testing"]["new"] == "Something pretended to change"
        assert ret[watch].full_return["__run_num__"] == 2
        assert ret[watch].comment == "Watch statement fired."


def test_watch_in_failure(state, state_tree):
    """
    test watch_in requisite when there is a failure
    """
    sls_contents = """
    return_changes:
      test.fail_with_changes:
        - watch_in:
          - test: watch_states

    watch_states:
      test.succeed_without_changes
    """
    fail = "test_|-return_changes_|-return_changes_|-fail_with_changes"
    watch = "test_|-watch_states_|-watch_states_|-succeed_without_changes"
    with pytest.helpers.temp_file("requisite.sls", sls_contents, state_tree):
        ret = state.sls("requisite")
        assert ret[fail].result is False
        assert (
            ret[watch].comment
            == "One or more requisite failed: requisite.return_changes"
        )


def test_requisites_watch_any(state, state_tree):
    """
    Call sls file containing several require_in and require.

    Ensure that some of them are failing and that the order is right.
    """
    if salt.utils.platform.is_windows():
        cmd_true = "exit"
        cmd_false = "exit /B 1"
    else:
        cmd_true = "true"
        cmd_false = "false"
    sls_contents = """
    A:
      cmd.wait:
        - name: '{cmd_true}'
        - watch_any:
          - cmd: B
          - cmd: C
          - cmd: D

    B:
      cmd.run:
        - name: '{cmd_true}'

    C:
      cmd.run:
        - name: '{cmd_false}'

    D:
      cmd.run:
        - name: '{cmd_true}'

    E:
      cmd.wait:
        - name: '{cmd_true}'
        - watch_any:
          - cmd: F
          - cmd: G
          - cmd: H

    F:
      cmd.run:
        - name: '{cmd_true}'

    G:
      cmd.run:
        - name: '{cmd_false}'

    H:
      cmd.run:
        - name: '{cmd_false}'
    """.format(
        cmd_true=cmd_true, cmd_false=cmd_false
    )
    expected_result = {
        f"cmd_|-A_|-{cmd_true}_|-wait": {
            "__run_num__": 4,
            "comment": f'Command "{cmd_true}" run',
            "result": True,
            "changes": True,
        },
        f"cmd_|-B_|-{cmd_true}_|-run": {
            "__run_num__": 0,
            "comment": f'Command "{cmd_true}" run',
            "result": True,
            "changes": True,
        },
        f"cmd_|-C_|-{cmd_false}_|-run": {
            "__run_num__": 1,
            "comment": f'Command "{cmd_false}" run',
            "result": False,
            "changes": True,
        },
        f"cmd_|-D_|-{cmd_true}_|-run": {
            "__run_num__": 2,
            "comment": f'Command "{cmd_true}" run',
            "result": True,
            "changes": True,
        },
        f"cmd_|-E_|-{cmd_true}_|-wait": {
            "__run_num__": 9,
            "comment": f'Command "{cmd_true}" run',
            "result": True,
            "changes": True,
        },
        f"cmd_|-F_|-{cmd_true}_|-run": {
            "__run_num__": 5,
            "comment": f'Command "{cmd_true}" run',
            "result": True,
            "changes": True,
        },
        f"cmd_|-G_|-{cmd_false}_|-run": {
            "__run_num__": 6,
            "comment": f'Command "{cmd_false}" run',
            "result": False,
            "changes": True,
        },
        f"cmd_|-H_|-{cmd_false}_|-run": {
            "__run_num__": 7,
            "comment": f'Command "{cmd_false}" run',
            "result": False,
            "changes": True,
        },
    }
    with pytest.helpers.temp_file("requisite.sls", sls_contents, state_tree):
        ret = state.sls("requisite")
        result = normalize_ret(ret.raw)
        assert result == expected_result


def test_requisites_watch_any_fail(state, state_tree):
    """
    Call sls file containing several require_in and require.

    Ensure that some of them are failing and that the order is right.
    """
    sls_contents = """
    A:
      cmd.wait:
        - name: 'true'
        - watch_any:
          - cmd: B
          - cmd: C

    B:
      cmd.run:
        - name: 'false'

    C:
      cmd.run:
        - name: 'false'
    """
    with pytest.helpers.temp_file("requisite.sls", sls_contents, state_tree):
        ret = state.sls("requisite")
        assert "One or more requisite failed" in ret["cmd_|-A_|-true_|-wait"].comment


def test_issue_30820_requisite_in_match_by_name(state, state_tree):
    """
    This tests the case where a requisite_in matches by name instead of ID

    See https://github.com/saltstack/salt/issues/30820 for more info
    """
    sls_contents = """
    bar state:
      cmd.wait:
        - name: 'echo bar'

    echo foo:
      cmd.run:
        - watch_in:
          - cmd: 'echo bar'
    """
    bar_state = "cmd_|-bar state_|-echo bar_|-wait"
    with pytest.helpers.temp_file("requisite.sls", sls_contents, state_tree):
        ret = state.sls("requisite")
        assert bar_state in ret
        assert ret[bar_state].comment == 'Command "echo bar" run'



import ipaddress
from datetime import datetime, timedelta, timezone

import pytest

import salt.exceptions
import salt.utils.x509 as x509
from tests.support.mock import ANY, Mock, patch

cryptography = pytest.importorskip(
    "cryptography", reason="Needs cryptography library", minversion="37.0"
)
cx509 = pytest.importorskip("cryptography.x509", reason="Needs cryptography library")
cprim = pytest.importorskip(
    "cryptography.hazmat.primitives", reason="Needs cryptography library"
)


@pytest.fixture
def single_pem():
    return """\
-----BEGIN RSA PRIVATE KEY-----
MIIEogIBAAKCAQEAu4FN2FAbh06p0SalHsN0uyvEK/pmxoJqgnPll+NIyUtyEKms
GXuRzZ6hF++a74dkk6/YbmyGAHECuS1WPm1iqaO7KASFEB3daeLg9r3jlFP4cxZI
n3UnEZF+GDYHdox00JyW0CLp6VUouXxc1gogkVGmeyi52b4LsOlGXdO8/8Nbcjhu
V9GZvdvvp6CwTLZabPni6nL/WWl5EwRqYnVTcX323tsObRliGsP6FYy1lQQLiy0A
PNIU+sxdQFeIdsR5FZ2JyquyKGD4REGrNGp2VuxzPrOmzr46AKTHRdDUaCedzUDO
wwK/sE4+b2DKEvyhYI1O5K5rERAFG4voZGwVTQIDAQABAoIBAGdu1LpWtljVk+fE
IaHuwB3f7r8zyi4HEyoTNLusrSIddDas4jrMZ4m9z6+chSNM1LaDcii3xNPJg28T
C1g0jxB1OXDBzYUudE1M0jxKU5gnGg6iZD1SKtMOJzjD1SoYIPhS6P63w8DrMSPg
7nVD8OM431VhCeSLaXeVtzNa8g9DirSHQD8zHHb3H7jnPJ4SeFKYv8URl3lsN84A
frP64nO64OGYgmIPU5IgVpKmK0rjXll4zNRxx5UwO2BglJ2lR6EIZ6oBA8ooCrPd
qOKrfiVTHoV9sKVwUtBvP3mqC1WZ6CWmX9EdnrSG0jgUZezezonHfcg8eBtoOPpR
0shrbwECgYEA6Ey8OGK3N3LNYewMuvv9gl66+TSbQY8Dh5iASHppvYFh6rWpdqo6
tRJORHbPn01fb2p48Pne8XBqDZNkG5u0gfX+v0yq2W5U0ezd0plP/uirLTuUm5if
hP525ShnHW6DYGsc26brCwRDINx4jYDqEiCq+1DcgoTavWBsUq+59aUCgYEAzqKc
TB5R/O6iwjCuJeIXddiIxHNFCwoaXSdorkCw2A7VJWUeh8SOpYuE//rfxW7pJbIS
kDc7rbdr/oyE7EYA6nIonWHJdXQ6V23SHARFWDX+LEn10OWOpSQnigSlcd1zFXLO
UDBVronmstqFriTugZfDAJZo7x6zsPfU/2X3IIkCgYBPaYweqoB/0BsuEof3lBWB
7+hzMOyyaLWIMTYJkO98/TIADsIz8tXG+M8Q0J0BlG2/pOJbXtA8MXXP1kcuuPfo
RbQkqYzub61HZnYefJLATcHW4LtYxcAisurqQ/mcMh9vYq6m2FUZmwdnwHblyOA7
+jb5WxdG9yvf+YqOacxkkQKBgE6AlKSOeFOBTbA80kxuIr+QrhUEPdy9z9pIGIrq
5MSQjgWQ7xJhaFgYM0UUyGK3ijfZ+Rd1BGUw5ARm2jDxP3PSPv/boK/QokGI5WPj
c3zZtmCZEJx2OcUfgS38KeaiXRBu91abplGS7mRQhKzuNvZg86KLgf4mSdoXrYIB
+OsRAoGANP4NjqBAEzsIFszGvaXmslyzfdWY6XI5N1GgzbEW/K7UECiPxWxmWcCm
n7mpiv/r8zkskgfAua84Fe8qIr2f6UjNeIzJ7LwI0zSB2uuiS4oP3xqCEyb+J3F4
bQdPnxzSwrf6edD2AmIT9L8IwiCYiplC+JvqSlqDP2pxIQbilmw=
-----END RSA PRIVATE KEY-----
    """


@pytest.fixture
def multi_pem():
    return """\
-----BEGIN CERTIFICATE-----
MIIDKDCCAhCgAwIBAgIIEmd4S33OHDEwDQYJKoZIhvcNAQELBQAwKzELMAkGA1UE
BhMCVVMxDTALBgNVBAMMBFRlc3QxDTALBgNVBAoMBFNhbHQwHhcNMjIxMTE1MTQw
OTQ0WhcNMzIxMTEyMTQwOTQ0WjAuMQswCQYDVQQGEwJVUzEQMA4GA1UEAwwHVGVz
dENydDENMAsGA1UECgwEU2FsdDCCASIwDQYJKoZIhvcNAQEBBQADggEPADCCAQoC
ggEBAOGTScvrjcEt6vsJcG9RUp6fKaDNDWZnJET0omanK9ZwaoGpJPp8UDYe/8AD
eI7N10wdyB4oDM9gRDjInBtdQO/PsrmKZF6LzqVFgLMxu2up+PHMi9z6B2P4esIA
zMu9PYxc9zH4HzLImHqscVD2HCabsjp9X134Af7hVY5NN/W/4qTP7uOM20wSG2TP
I6+BtA9VyPbEPMPRzXzrqc45rVYe6kb2bT84GE93Vcu/e5JZ/k2AKD8Hoa2cxLPs
TLq5igl+D+k+dfUtiABiKPvVQiYBsD1fyHDn2m7B6pCgvrGqHjsoAKufgFnXy6PJ
Rg7nvQfaxSiusM5s+VS+fjlvgwsCAwEAAaNNMEswCQYDVR0TBAIwADAdBgNVHQ4E
FgQUXPLx9FMpI57uQFqSo7QqLvWmi2cwHwYDVR0jBBgwFoAUXPLx9FMpI57uQFqS
o7QqLvWmi2cwDQYJKoZIhvcNAQELBQADggEBAFkYke/32PEmCDcs8YFDp5PDVwhv
8ZlfCtgRChaPcwoc+QpBeNhUEMhQ/v77Ojj6VNXzNDP7X1MCc7432Xo3VHCiCI8c
CMVNiyDWJNlEEll3eUIguNWRnkIfsCeHVs5/76M9tgwcrPOltQliI4UJ7Wy5pnXA
ywQFn43TqzNdU7HRIdX7QRUHeMOpe7Y4o7vjSOseQJcL2pQadho/BqK5rWuwZ8x/
Jsl40YjjXyLOnOCUYYpC13KC8iHOZKHVvA8eDtiuuLSgIk9DBD/ClU4cMCF2mvuQ
dtV9rM6byvPUJaIxKznODAUbZPGuFI/FWP9VheO17s7mDs/B12/UKXcTkvs=
-----END CERTIFICATE-----

-----BEGIN CERTIFICATE-----
MIIDODCCAiCgAwIBAgIIbfpgqP0VGPgwDQYJKoZIhvcNAQELBQAwKzELMAkGA1UE
BhMCVVMxDTALBgNVBAMMBFRlc3QxDTALBgNVBAoMBFNhbHQwHhcNMjIxMTE1MTQw
NDMzWhcNMzIxMTEyMTQwNDMzWjArMQswCQYDVQQGEwJVUzENMAsGA1UEAwwEVGVz
dDENMAsGA1UECgwEU2FsdDCCASIwDQYJKoZIhvcNAQEBBQADggEPADCCAQoCggEB
AOGTScvrjcEt6vsJcG9RUp6fKaDNDWZnJET0omanK9ZwaoGpJPp8UDYe/8ADeI7N
10wdyB4oDM9gRDjInBtdQO/PsrmKZF6LzqVFgLMxu2up+PHMi9z6B2P4esIAzMu9
PYxc9zH4HzLImHqscVD2HCabsjp9X134Af7hVY5NN/W/4qTP7uOM20wSG2TPI6+B
tA9VyPbEPMPRzXzrqc45rVYe6kb2bT84GE93Vcu/e5JZ/k2AKD8Hoa2cxLPsTLq5
igl+D+k+dfUtiABiKPvVQiYBsD1fyHDn2m7B6pCgvrGqHjsoAKufgFnXy6PJRg7n
vQfaxSiusM5s+VS+fjlvgwsCAwEAAaNgMF4wDwYDVR0TBAgwBgEB/wIBATALBgNV
HQ8EBAMCAQYwHQYDVR0OBBYEFFzy8fRTKSOe7kBakqO0Ki71potnMB8GA1UdIwQY
MBaAFFzy8fRTKSOe7kBakqO0Ki71potnMA0GCSqGSIb3DQEBCwUAA4IBAQBZS4MP
fXYPoGZ66seM+0eikScZHirbRe8vHxHkujnTBUjQITKm86WeQgeBCD2pobgBGZtt
5YFozM4cERqY7/1BdemUxFvPmMFFznt0TM5w+DfGWVK8un6SYwHnmBbnkWgX4Srm
GsL0HHWxVXkGnFGFk6Sbo3vnN7CpkpQTWFqeQQ5rHOw91pt7KnNZwc6I3ZjrCUHJ
+UmKKrga16a4Q+8FBpYdphQU609npo/0zuaE6FyiJYlW3tG+mlbbNgzY/+eUaxt2
9Bp9mtA+Hkox551Mfpq45Oi+ehwMt0xjZCjuFCM78oiUdHCGO+EmcT7ogiYALiOF
LN1w5sybsYwIw6QN
-----END CERTIFICATE-----
    """


def test_split_pems_single(single_pem):
    res = x509.split_pems(single_pem)
    assert len(res) == 1
    assert res[0].startswith(b"-----BEGIN RSA PRIVATE KEY-----\n")
    assert res[0].endswith(b"-----END RSA PRIVATE KEY-----\n")
    assert len(res[0].splitlines()) == 27


def test_split_pems_multi(multi_pem):
    res = x509.split_pems(multi_pem)
    assert len(res) == 2
    for x in res:
        assert x.startswith(b"-----BEGIN CERTIFICATE-----\n")
        assert x.endswith(b"-----END CERTIFICATE-----\n")
    assert len(res[0].splitlines()) == 19
    assert len(res[1].splitlines()) == 20


def test_split_pems_garbage_between(single_pem):
    garbage_pem = (
        single_pem
        + "\nI like turtles\nIntroduce a little anarchy. Upset the established order, and everything becomes chaos. I'm an agent of chaos.\n"
        + single_pem
    )
    res = x509.split_pems(garbage_pem)
    assert len(res) == 2
    for x in res:
        assert x.startswith(b"-----BEGIN RSA PRIVATE KEY-----\n")
        assert x.endswith(b"-----END RSA PRIVATE KEY-----\n")
        assert len(x.splitlines()) == 27


class TestCreateExtension:
    @pytest.fixture
    def aki(self):
        with patch("cryptography.x509.AuthorityKeyIdentifier", autospec=True) as ext:
            yield ext

    @pytest.fixture
    def ca_crt(self):
        ca = Mock(spec=cx509.Certificate)
        return ca

    @pytest.mark.parametrize(
        "val,expected,critical",
        [
            ("critical,CA:FALSE", (False, None), True),
            ("critical, CA:TRUE, pathlen:2", (True, 2), True),
            ("CA:TRUE", (True, None), False),
            ({"ca": False, "critical": True}, (False, None), True),
            ({"ca": True, "pathlen": 3}, (True, 3), False),
        ],
    )
    def test_create_basic_constraints(self, val, expected, critical):
        with patch("cryptography.x509.BasicConstraints", autospec=True) as ext:
            _, crit = x509._create_extension("basicConstraints", val)
            assert crit == critical
            ext.assert_called_once_with(*expected)

    @pytest.mark.parametrize(
        "val,expected,critical",
        [
            (
                "digitalSignature, nonRepudiation, keyEncipherment, dataEncipherment, keyAgreement, keyCertSign, cRLSign, encipherOnly, decipherOnly",
                {
                    "digital_signature": True,
                    "content_commitment": True,
                    "key_encipherment": True,
                    "data_encipherment": True,
                    "key_agreement": True,
                    "key_cert_sign": True,
                    "crl_sign": True,
                    "encipher_only": True,
                    "decipher_only": True,
                },
                False,
            ),
            (
                "critical, keyCertSign, cRLSign",
                {
                    "digital_signature": False,
                    "content_commitment": False,
                    "key_encipherment": False,
                    "data_encipherment": False,
                    "key_agreement": False,
                    "key_cert_sign": True,
                    "crl_sign": True,
                    "encipher_only": False,
                    "decipher_only": False,
                },
                True,
            ),
            (
                ["critical", "digitalSignature"],
                {
                    "digital_signature": True,
                    "content_commitment": False,
                    "key_encipherment": False,
                    "data_encipherment": False,
                    "key_agreement": False,
                    "key_cert_sign": False,
                    "crl_sign": False,
                    "encipher_only": False,
                    "decipher_only": False,
                },
                True,
            ),
            (
                ["nonRepudiation"],
                {
                    "digital_signature": False,
                    "content_commitment": True,
                    "key_encipherment": False,
                    "data_encipherment": False,
                    "key_agreement": False,
                    "key_cert_sign": False,
                    "crl_sign": False,
                    "encipher_only": False,
                    "decipher_only": False,
                },
                False,
            ),
        ],
    )
    def test_create_key_usage(self, val, expected, critical):
        with patch("cryptography.x509.KeyUsage", autospec=True) as ext:
            _, crit = x509._create_extension("keyUsage", val)
            assert crit == critical
            ext.assert_called_once_with(**expected)

    @pytest.mark.parametrize(
        "val,expected,critical",
        [
            (
                "critical,serverAuth,clientAuth,codeSigning,emailProtection,timeStamping,OCSPSigning,msSmartcardLogin,pkInitKDC,ipsecIKE,msCodeInd,msCodeCom,msCTLSign,msEFS",
                list(x509.EXTENDED_KEY_USAGE_OID.values()),
                True,
            ),
            ("serverAuth", [x509.EXTENDED_KEY_USAGE_OID["serverAuth"]], False),
            (
                "timeStamping,1.2.3.4",
                [
                    x509.EXTENDED_KEY_USAGE_OID["timeStamping"],
                    cx509.ObjectIdentifier("1.2.3.4"),
                ],
                False,
            ),
            (
                ["critical", "OCSPSigning", "msCodeCom", "2.3.4.55"],
                [
                    x509.EXTENDED_KEY_USAGE_OID["OCSPSigning"],
                    x509.EXTENDED_KEY_USAGE_OID["msCodeCom"],
                    cx509.ObjectIdentifier("2.3.4.55"),
                ],
                True,
            ),
            (["clientAuth"], [x509.EXTENDED_KEY_USAGE_OID["clientAuth"]], False),
        ],
    )
    def test_create_extended_key_usage(self, val, expected, critical):
        with patch("cryptography.x509.ExtendedKeyUsage", autospec=True) as ext:
            _, crit = x509._create_extension("extendedKeyUsage", val)
            assert crit == critical
            ext.assert_called_once_with(expected)

    @pytest.mark.parametrize(
        "val,expected",
        [
            ("hash", None),
            (
                "BB:AF:7E:02:3D:FA:A6:F1:3C:84:8E:AD:EE:38:98:EC:D9:32:32:D4",
                b"\xbb\xaf~\x02=\xfa\xa6\xf1<\x84\x8e\xad\xee8\x98\xec\xd922\xd4",
            ),
            (
                "bbaf7e023dfaa6f13c848eadee3898ecd93232d4",
                b"\xbb\xaf~\x02=\xfa\xa6\xf1<\x84\x8e\xad\xee8\x98\xec\xd922\xd4",
            ),
        ],
    )
    def test_create_subject_key_identifier(self, val, expected):
        with patch("cryptography.x509.SubjectKeyIdentifier", autospec=True) as ext:
            _, crit = x509._create_extension(
                "subjectKeyIdentifier", val, subject_pubkey="testpub"
            )
            assert crit is False
            if val == "hash":
                ext.from_public_key.assert_called_once_with("testpub")
            else:
                ext.from_public_key.assert_not_called()
                ext.assert_called_once_with(expected)

    def test_create_authority_key_identifier_from_ski(self, aki, ca_crt):
        ca_crt.extensions.get_extension_for_class.return_value.value.digest = (
            b"\xde\xad\xbe\xef"
        )
        _, crit = x509._create_extension(
            "authorityKeyIdentifier", "keyid:always", ca_crt=ca_crt
        )
        assert crit is False
        aki.assert_called_once_with(
            key_identifier=b"\xde\xad\xbe\xef",
            authority_cert_issuer=None,
            authority_cert_serial_number=None,
        )

    def test_create_authority_key_identifier_from_pubkey(self, aki, ca_crt):
        def raise_notfound(*args, **kwargs):
            raise cx509.ExtensionNotFound("f", cx509.ObjectIdentifier("1.2.3.4"))

        ca_crt.extensions.get_extension_for_class.side_effect = raise_notfound
        aki.from_issuer_public_key.return_value.key_identifier = b"\xde\xad\xbe\xef"
        _, crit = x509._create_extension(
            "authorityKeyIdentifier", "keyid:always", ca_crt=ca_crt
        )
        assert crit is False
        aki.assert_called_once_with(
            key_identifier=b"\xde\xad\xbe\xef",
            authority_cert_issuer=None,
            authority_cert_serial_number=None,
        )

    def test_create_authority_key_identifier_from_ca_pub(self, aki):
        aki.from_issuer_public_key.return_value.key_identifier = b"\xde\xad\xbe\xef"
        _, crit = x509._create_extension(
            "authorityKeyIdentifier", "keyid:always", ca_pub="testpub"
        )
        assert crit is False
        aki.assert_called_once_with(
            key_identifier=b"\xde\xad\xbe\xef",
            authority_cert_issuer=None,
            authority_cert_serial_number=None,
        )

    @pytest.mark.parametrize(
        "ca_crt,ca_pub",
        [
            (Mock(), None),
            (None, Mock()),
            (Mock(), Mock()),
        ],
    )
    def test_create_authority_key_identifier_always(self, aki, ca_crt, ca_pub):
        aki.from_issuer_public_key.side_effect = RuntimeError
        if ca_crt is not None:
            ca_crt.extensions.get_extension_for_class.side_effect = RuntimeError
        elif ca_pub is not None:
            ca_pub.side_effect = RuntimeError
        with pytest.raises(salt.exceptions.CommandExecutionError):
            x509._create_extension(
                "authorityKeyIdentifier", "keyid:always", ca_crt=ca_crt, ca_pub=ca_pub
            )

    def test_create_authority_key_identifier_issuer(self, aki, ca_crt):
        ca_crt.issuer = "testissuer"
        ca_crt.serial_number = 1337
        with patch("cryptography.x509.DirectoryName") as dirname:
            _, crit = x509._create_extension(
                "authorityKeyIdentifier", "issuer", ca_crt=ca_crt
            )
            assert crit is False
            dirname.assert_called_once_with(ca_crt.issuer)
            aki.assert_called_once_with(
                key_identifier=None,
                authority_cert_issuer=ANY,
                authority_cert_serial_number=1337,
            )

    def test_create_authority_key_identifier_issuer_always(self, aki, ca_crt):
        with patch("cryptography.x509.DirectoryName") as dirname:
            dirname.side_effect = ValueError
            with pytest.raises(salt.exceptions.CommandExecutionError):
                x509._create_extension(
                    "authorityKeyIdentifier", "issuer:always", ca_crt=ca_crt
                )

    def test_create_authority_key_identifier_from_both(self, aki, ca_crt):
        ca_crt.issuer = "testissuer"
        ca_crt.serial_number = 1337
        ca_crt.extensions.get_extension_for_class.return_value.value.digest = (
            b"\xde\xad\xbe\xef"
        )
        with patch("cryptography.x509.DirectoryName"):
            _, crit = x509._create_extension(
                "authorityKeyIdentifier", "keyid:always,issuer:always", ca_crt=ca_crt
            )
            assert crit is False
            aki.assert_called_once_with(
                key_identifier=b"\xde\xad\xbe\xef",
                authority_cert_issuer=ANY,
                authority_cert_serial_number=1337,
            )

    def test_create_authority_key_identifier_from_both_issuer_fail(self, aki, ca_crt):
        ca_crt.issuer = "testissuer"
        ca_crt.serial_number = 1337
        ca_crt.extensions.get_extension_for_class.return_value.value.digest = (
            b"\xde\xad\xbe\xef"
        )
        with patch("cryptography.x509.DirectoryName") as dirname:
            dirname.side_effect = ValueError
            _, crit = x509._create_extension(
                "authorityKeyIdentifier", "keyid:always,issuer", ca_crt=ca_crt
            )
            assert crit is False
            aki.assert_called_once_with(
                key_identifier=b"\xde\xad\xbe\xef",
                authority_cert_issuer=None,
                authority_cert_serial_number=None,
            )

    @pytest.mark.parametrize(
        "val,ca_sub,expected,critical",
        [
            (
                ["email:ca@example.com", "dns:example.com"],
                None,
                [
                    cx509.RFC822Name("ca@example.com"),
                    cx509.DNSName("example.com"),
                ],
                False,
            ),
            (
                ["issuer:copy"],
                [
                    cx509.RFC822Name("me@example.com"),
                    cx509.DNSName("example.com"),
                ],
                [
                    cx509.RFC822Name("me@example.com"),
                    cx509.DNSName("example.com"),
                ],
                False,
            ),
            (
                ["critical", "issuer:copy", "dns:example.io"],
                [
                    cx509.RFC822Name("ca@example.com"),
                    cx509.DNSName("example.com"),
                ],
                [
                    cx509.RFC822Name("ca@example.com"),
                    cx509.DNSName("example.com"),
                    cx509.DNSName("example.io"),
                ],
                True,
            ),
            (
                "critical,issuer:copy,dns:example.io",
                [
                    cx509.RFC822Name("ca@example.com"),
                    cx509.DNSName("example.com"),
                ],
                [
                    cx509.RFC822Name("ca@example.com"),
                    cx509.DNSName("example.com"),
                    cx509.DNSName("example.io"),
                ],
                True,
            ),
            (
                "DNS:salt.ca",
                None,
                [
                    cx509.DNSName("salt.ca"),
                ],
                False,
            ),
        ],
    )
    @pytest.mark.parametrize(
        "tgt,extname",
        [
            ("IssuerAlternativeName", "issuerAltName"),
            ("CertificateIssuer", "certificateIssuer"),
        ],
    )
    def test_create_issuer_alt_name(
        self, val, ca_sub, expected, critical, ca_crt, tgt, extname
    ):
        ca_crt.extensions.get_extension_for_class.return_value._general_names._general_names = (
            ca_sub
        )
        with patch(f"cryptography.x509.{tgt}", autospec=True) as ext:
            res, crit = x509._create_extension(extname, val, ca_crt=ca_crt)
            assert crit == critical
            ext.assert_called_once_with(expected)

    @pytest.mark.parametrize(
        "val,expected",
        [
            (
                "OCSP;URI:http://ocsp.example.com/,caIssuers;URI:http://myca.example.com/ca.cer",
                [
                    cx509.AccessDescription(
                        access_method=x509.ACCESS_OID["OCSP"],
                        access_location=cx509.UniformResourceIdentifier(
                            "http://ocsp.example.com/"
                        ),
                    ),
                    cx509.AccessDescription(
                        access_method=x509.ACCESS_OID["caIssuers"],
                        access_location=cx509.UniformResourceIdentifier(
                            "http://myca.example.com/ca.cer"
                        ),
                    ),
                ],
            ),
            (
                {
                    "OCSP": "URI:http://ocsp.example.com/",
                },
                [
                    cx509.AccessDescription(
                        access_method=x509.ACCESS_OID["OCSP"],
                        access_location=cx509.UniformResourceIdentifier(
                            "http://ocsp.example.com/"
                        ),
                    ),
                ],
            ),
            (
                [
                    {"OCSP": "URI:http://ocsp.example.com/"},
                    {"OCSP": "URI:http://ocsp2.example.com/"},
                ],
                [
                    cx509.AccessDescription(
                        access_method=x509.ACCESS_OID["OCSP"],
                        access_location=cx509.UniformResourceIdentifier(
                            "http://ocsp.example.com/"
                        ),
                    ),
                    cx509.AccessDescription(
                        access_method=x509.ACCESS_OID["OCSP"],
                        access_location=cx509.UniformResourceIdentifier(
                            "http://ocsp2.example.com/"
                        ),
                    ),
                ],
            ),
        ],
    )
    def test_create_authority_info_access(self, val, expected):
        with patch(
            "cryptography.x509.AuthorityInformationAccess", autospec=True
        ) as ext:
            res, crit = x509._create_extension("authorityInfoAccess", val)
            assert crit is False
            ext.assert_called_once_with(expected)

    @pytest.mark.parametrize(
        "val,expected,critical",
        [
            (
                ["email:user@example.com", "dns:example.com"],
                [
                    cx509.RFC822Name("user@example.com"),
                    cx509.DNSName("example.com"),
                ],
                False,
            ),
            (
                ["critical", "dns:example.io"],
                [cx509.DNSName("example.io")],
                True,
            ),
            (
                "critical,dns:example.io,email:hello@example.io",
                [
                    cx509.DNSName("example.io"),
                    cx509.RFC822Name("hello@example.io"),
                ],
                True,
            ),
        ],
    )
    def test_create_subject_alt_name(self, val, expected, critical):
        with patch("cryptography.x509.SubjectAlternativeName", autospec=True) as ext:
            res, crit = x509._create_extension("subjectAltName", val)
            assert crit == critical
            ext.assert_called_once_with(expected)

    @pytest.mark.parametrize(
        "val,expected,critical",
        [
            (
                "URI:http://example.com/myca.crl, URI:http://example.org/my.crl",
                [
                    cx509.DistributionPoint(
                        full_name=[
                            cx509.UniformResourceIdentifier(
                                "http://example.com/myca.crl"
                            )
                        ],
                        relative_name=None,
                        reasons=None,
                        crl_issuer=None,
                    ),
                    cx509.DistributionPoint(
                        full_name=[
                            cx509.UniformResourceIdentifier("http://example.org/my.crl")
                        ],
                        relative_name=None,
                        reasons=None,
                        crl_issuer=None,
                    ),
                ],
                False,
            ),
            (
                "critical,URI:http://example.com/myca.crl",
                [
                    cx509.DistributionPoint(
                        full_name=[
                            cx509.UniformResourceIdentifier(
                                "http://example.com/myca.crl"
                            )
                        ],
                        relative_name=None,
                        reasons=None,
                        crl_issuer=None,
                    )
                ],
                True,
            ),
            (
                [
                    {
                        "fullname": [
                            "URI:http://example.com/myca.crl",
                            "URI:http://example.org/my.crl",
                        ]
                    }
                ],
                [
                    cx509.DistributionPoint(
                        full_name=[
                            cx509.UniformResourceIdentifier(
                                "http://example.com/myca.crl"
                            ),
                            cx509.UniformResourceIdentifier(
                                "http://example.org/my.crl"
                            ),
                        ],
                        relative_name=None,
                        reasons=None,
                        crl_issuer=None,
                    )
                ],
                False,
            ),
            (
                [
                    {
                        "fullname": "URI:http://example.com/myca.crl",
                        "crlissuer": "DNS:example.org",
                        "reasons": ["keyCompromise"],
                    }
                ],
                [
                    cx509.DistributionPoint(
                        full_name=[
                            cx509.UniformResourceIdentifier(
                                "http://example.com/myca.crl"
                            )
                        ],
                        relative_name=None,
                        reasons=frozenset([cx509.ReasonFlags("keyCompromise")]),
                        crl_issuer=[cx509.DNSName("example.org")],
                    )
                ],
                False,
            ),
            pytest.param(
                [
                    "critical",
                    {
                        "relativename": "OU=foo+CN=Smith",
                        "crlissuer": ["DNS:example.org"],
                    },
                ],
                [
                    cx509.DistributionPoint(
                        full_name=None,
                        relative_name=cx509.RelativeDistinguishedName(
                            [
                                cx509.NameAttribute(
                                    cx509.ObjectIdentifier("2.5.4.11"), value="foo"
                                ),
                                cx509.NameAttribute(
                                    cx509.ObjectIdentifier("2.5.4.3"), value="Smith"
                                ),
                            ]
                        ),
                        reasons=None,
                        crl_issuer=[cx509.DNSName("example.org")],
                    )
                ],
                True,
            ),
        ],
    )
    @pytest.mark.parametrize(
        "tgt,extname",
        [
            ("CRLDistributionPoints", "crlDistributionPoints"),
            ("FreshestCRL", "freshestCRL"),
        ],
    )
    def test_create_crl_distribution_points_freshest_crl(
        self, val, expected, critical, tgt, extname
    ):
        with patch(f"cryptography.x509.{tgt}", autospec=True) as ext:
            res, crit = x509._create_extension(extname, val)
            if tgt == "FreshestCRL":
                assert crit is False
            else:
                assert crit == critical
            ext.assert_called_once_with(expected)

    @pytest.mark.parametrize(
        "val,expected,critical",
        [
            (
                {
                    "critical": True,
                    "fullname": [
                        "URI:http://example.com/myca.crl",
                        "URI:http://example.org/my.crl",
                    ],
                },
                {
                    "full_name": [
                        cx509.UniformResourceIdentifier("http://example.com/myca.crl"),
                        cx509.UniformResourceIdentifier("http://example.org/my.crl"),
                    ],
                    "relative_name": None,
                    "only_some_reasons": None,
                    "only_contains_user_certs": False,
                    "only_contains_ca_certs": False,
                    "only_contains_attribute_certs": False,
                    "indirect_crl": False,
                },
                True,
            ),
            (
                {
                    "fullname": "URI:http://example.com/myca.crl",
                    "onlysomereasons": ["keyCompromise"],
                    "onlyuser": True,
                    "onlyCA": False,
                    "onlyAA": False,
                    "indirectCRL": False,
                },
                {
                    "full_name": [
                        cx509.UniformResourceIdentifier("http://example.com/myca.crl")
                    ],
                    "relative_name": None,
                    "only_some_reasons": frozenset(
                        [cx509.ReasonFlags("keyCompromise")]
                    ),
                    "only_contains_user_certs": True,
                    "only_contains_ca_certs": False,
                    "only_contains_attribute_certs": False,
                    "indirect_crl": False,
                },
                False,
            ),
        ],
    )
    def test_create_issuing_distribution_point(self, val, expected, critical):
        with patch("cryptography.x509.IssuingDistributionPoint", autospec=True) as ext:
            res, crit = x509._create_extension("issuingDistributionPoint", val)
            assert crit == critical
            ext.assert_called_once_with(**expected)

    @pytest.mark.parametrize(
        "val,expected,critical",
        [
            (
                "critical, 1.2.4.5, 1.1.3.4",
                [
                    cx509.PolicyInformation(
                        policy_identifier=cx509.ObjectIdentifier("1.2.4.5"),
                        policy_qualifiers=None,
                    ),
                    cx509.PolicyInformation(
                        policy_identifier=cx509.ObjectIdentifier("1.1.3.4"),
                        policy_qualifiers=None,
                    ),
                ],
                True,
            ),
            (
                {
                    "critical": True,
                    "1.2.3.4.5": ["https://my.ca.com/pratice_statement"],
                },
                [
                    cx509.PolicyInformation(
                        policy_identifier=cx509.ObjectIdentifier("1.2.3.4.5"),
                        policy_qualifiers=["https://my.ca.com/pratice_statement"],
                    )
                ],
                True,
            ),
            (
                {
                    "1.2.3.4.5": [
                        "https://my.ca.com/pratice_statement",
                        {
                            "organization": "myorg",
                            "noticeNumbers": [1, 2, 3],
                            "text": "mytext",
                        },
                    ]
                },
                [
                    cx509.PolicyInformation(
                        policy_identifier=cx509.ObjectIdentifier("1.2.3.4.5"),
                        policy_qualifiers=[
                            "https://my.ca.com/pratice_statement",
                            cx509.UserNotice(
                                notice_reference=cx509.NoticeReference(
                                    organization="myorg", notice_numbers=[1, 2, 3]
                                ),
                                explicit_text="mytext",
                            ),
                        ],
                    )
                ],
                False,
            ),
        ],
    )
    def test_create_certificate_policies(self, val, expected, critical):
        with patch("cryptography.x509.CertificatePolicies", autospec=True) as ext:
            res, crit = x509._create_extension("certificatePolicies", val)
            assert crit == critical
            ext.assert_called_once_with(expected)

    @pytest.mark.parametrize(
        "val,expected,critical",
        [
            (
                "requireExplicitPolicy:3",
                {"require_explicit_policy": 3, "inhibit_policy_mapping": None},
                False,
            ),
            (
                "critical,requireExplicitPolicy:2,inhibitPolicyMapping:1",
                {"require_explicit_policy": 2, "inhibit_policy_mapping": 1},
                True,
            ),
            (
                {
                    "critical": True,
                    "requireExplicitPolicy": 4,
                    "inhibitPolicyMapping": 2,
                },
                {"require_explicit_policy": 4, "inhibit_policy_mapping": 2},
                True,
            ),
        ],
    )
    def test_create_policy_constraints(self, val, expected, critical):
        with patch("cryptography.x509.PolicyConstraints", autospec=True) as ext:
            res, crit = x509._create_extension("policyConstraints", val)
            assert crit == critical
            ext.assert_called_once_with(**expected)

    @pytest.mark.parametrize(
        "val,expected,critical",
        [
            (1, 1, False),
            ("critical, 1", 1, True),
            ({"critical": True, "value": 1}, 1, True),
        ],
    )
    def test_create_inhibit_any_policy(self, val, expected, critical):
        with patch("cryptography.x509.InhibitAnyPolicy", autospec=True) as ext:
            res, crit = x509._create_extension("inhibitAnyPolicy", val)
            assert crit == critical
            ext.assert_called_once_with(expected)

    @pytest.mark.parametrize(
        "val,excluded,permitted,critical",
        [
            (
                "critical,permitted;IP:192.168.0.0/255.255.0.0,permitted;email:.example.com,excluded;email:.com",
                [cx509.RFC822Name(".com")],
                [
                    cx509.IPAddress(ipaddress.ip_network("192.168.0.0/16")),
                    cx509.RFC822Name(".example.com"),
                ],
                True,
            ),
            (
                {
                    "critical": True,
                    "permitted": ["IP:192.168.0.0/255.255.0.0", "email:.example.com"],
                    "excluded": ["email:.com"],
                },
                [cx509.RFC822Name(".com")],
                [
                    cx509.IPAddress(ipaddress.ip_network("192.168.0.0/16")),
                    cx509.RFC822Name(".example.com"),
                ],
                True,
            ),
        ],
    )
    def test_create_name_constraints(self, val, excluded, permitted, critical):
        with patch("cryptography.x509.NameConstraints", autospec=True) as ext:
            res, crit = x509._create_extension("nameConstraints", val)
            assert crit == critical
            ext.assert_called_once_with(
                permitted_subtrees=permitted, excluded_subtrees=excluded
            )

    def test_create_name_constraints_requires_at_least_one_definition(self):
        with patch("cryptography.x509.NameConstraints", autospec=True):
            with pytest.raises(salt.exceptions.SaltInvocationError):
                x509._create_extension("nameConstraints", {"permitted": []})

    @pytest.mark.parametrize(
        "val,critical", [([], False), ("critical", True), (["critical"], True)]
    )
    def test_create_no_check(self, val, critical):
        with patch("cryptography.x509.OCSPNoCheck", autospec=True):
            res, crit = x509._create_extension("noCheck", val)
            assert crit == critical

    @pytest.mark.parametrize(
        "val,expected,critical",
        [
            (
                "critical, status_request, status_request_v2",
                [
                    cx509.TLSFeatureType.status_request,
                    cx509.TLSFeatureType.status_request_v2,
                ],
                True,
            ),
            (
                ["critical", "status_request"],
                [cx509.TLSFeatureType.status_request],
                True,
            ),
        ],
    )
    def test_create_tlsfeature(self, val, expected, critical):
        with patch("cryptography.x509.TLSFeature", autospec=True) as ext:
            res, crit = x509._create_extension("tlsfeature", val)
            assert crit == critical
            ext.assert_called_once_with(expected)

    def test_create_crl_number(self):
        with patch("cryptography.x509.CRLNumber", autospec=True) as ext:
            res, crit = x509._create_extension("cRLNumber", 3)
            assert crit is False
            ext.assert_called_once_with(3)

    @pytest.mark.parametrize(
        "val,expected,critical",
        [
            (
                "critical, 3",
                3,
                True,
            ),
            (
                "3",
                3,
                False,
            ),
            (
                3,
                3,
                False,
            ),
        ],
    )
    def test_create_delta_crl_indicator(self, val, expected, critical):
        with patch("cryptography.x509.DeltaCRLIndicator", autospec=True) as ext:
            res, crit = x509._create_extension("deltaCRLIndicator", val)
            assert crit == critical
            ext.assert_called_once_with(expected)

    @pytest.mark.parametrize(
        "val,expected,critical",
        [
            (
                "critical, keyCompromise",
                cx509.ReasonFlags("keyCompromise"),
                True,
            ),
            (
                "aACompromise",
                cx509.ReasonFlags("aACompromise"),
                False,
            ),
            (
                {"critical": True, "superseded": True},
                cx509.ReasonFlags("superseded"),
                True,
            ),
            (
                ["critical", "privilegeWithdrawn"],
                cx509.ReasonFlags("privilegeWithdrawn"),
                True,
            ),
        ],
    )
    def test_create_crl_reason(self, val, expected, critical):
        with patch("cryptography.x509.CRLReason", autospec=True) as ext:
            res, crit = x509._create_extension("CRLReason", val)
            assert crit == critical
            ext.assert_called_once_with(expected)

    @pytest.mark.parametrize(
        "val,expected,critical",
        [
            (
                "critical, 2022-10-11 13:37:42",
                datetime.strptime("2022-10-11 13:37:42", "%Y-%m-%d %H:%M:%S"),
                True,
            ),
            (
                "2022-10-11 13:37:42",
                datetime.strptime("2022-10-11 13:37:42", "%Y-%m-%d %H:%M:%S"),
                False,
            ),
        ],
    )
    def test_create_invalidity_date(self, val, expected, critical):
        with patch("cryptography.x509.InvalidityDate", autospec=True) as ext:
            res, crit = x509._create_extension("invalidityDate", val)
            assert crit == critical
            ext.assert_called_once_with(expected)


@pytest.mark.parametrize(
    "inpt,cls,parsed",
    [
        (("email", "me@example.com"), cx509.RFC822Name, "me@example.com"),
        (("email", ".example.com"), cx509.RFC822Name, ".example.com"),
        (
            ("email", "me@überexample.com"),
            cx509.RFC822Name,
            "me@xn--berexample-8db.com",
        ),
        (
            ("URI", "https://www.example.com"),
            cx509.UniformResourceIdentifier,
            "https://www.example.com",
        ),
        (
            ("URI", "https://www.überexample.com"),
            cx509.UniformResourceIdentifier,
            "https://www.xn--berexample-8db.com",
        ),
        (("URI", "some/path/only"), cx509.UniformResourceIdentifier, "some/path/only"),
        (("DNS", "example.com"), cx509.DNSName, "example.com"),
        (("DNS", "überexample.com"), cx509.DNSName, "xn--berexample-8db.com"),
        (("DNS", "*.überexample.com"), cx509.DNSName, "*.xn--berexample-8db.com"),
        (("DNS", ".überexample.com"), cx509.DNSName, ".xn--berexample-8db.com"),
        (
            ("DNS", "γνῶθι.σεαυτόν.gr"),
            cx509.DNSName,
            "xn--oxakdo9327a.xn--mxahzvhf4c.gr",
        ),
        (("RID", "1.2.3.4"), cx509.RegisteredID, cx509.ObjectIdentifier("1.2.3.4")),
        (
            ("IP", "13.37.13.37"),
            cx509.IPAddress,
            ipaddress.ip_address("13.37.13.37"),
        ),
        (
            ("IP", "13.37.13.0/24"),
            cx509.IPAddress,
            ipaddress.ip_network("13.37.13.0/24"),
        ),
        (
            ("IP", "13.37.13.0/255.255.255.0"),
            cx509.IPAddress,
            ipaddress.ip_network("13.37.13.0/255.255.255.0"),
        ),
        (
            ("IP", "2001:0db8:85a3:0000:0000:8a2e:0370:7334"),
            cx509.IPAddress,
            ipaddress.ip_address("2001:0db8:85a3:0000:0000:8a2e:0370:7334"),
        ),
        (
            ("IP", "2001:db8:abcd:0012::0/64"),
            cx509.IPAddress,
            ipaddress.ip_network("2001:db8:abcd:0012::0/64"),
        ),
        pytest.param(
            (
                "dirName",
                "CN=mysite.com,O=My Company,L=San Francisco,ST=California,C=US",
            ),
            cx509.Name,
            [
                cx509.RelativeDistinguishedName(
                    [cx509.NameAttribute(cx509.ObjectIdentifier("2.5.4.6"), value="US")]
                ),
                cx509.RelativeDistinguishedName(
                    [
                        cx509.NameAttribute(
                            cx509.ObjectIdentifier("2.5.4.8"), value="California"
                        )
                    ]
                ),
                cx509.RelativeDistinguishedName(
                    [
                        cx509.NameAttribute(
                            cx509.ObjectIdentifier("2.5.4.7"), value="San Francisco"
                        )
                    ]
                ),
                cx509.RelativeDistinguishedName(
                    [
                        cx509.NameAttribute(
                            cx509.ObjectIdentifier("2.5.4.10"), value="My Company"
                        )
                    ]
                ),
                cx509.RelativeDistinguishedName(
                    [
                        cx509.NameAttribute(
                            cx509.ObjectIdentifier("2.5.4.3"), value="mysite.com"
                        )
                    ]
                ),
            ],
        ),
        (
            (
                "dirName",
                {
                    "C": "US",
                    "ST": "California",
                    "L": "San Francisco",
                    "O": "My Company",
                    "CN": "mysite.com",
                },
            ),
            cx509.Name,
            [
                cx509.RelativeDistinguishedName(
                    [cx509.NameAttribute(cx509.ObjectIdentifier("2.5.4.6"), value="US")]
                ),
                cx509.RelativeDistinguishedName(
                    [
                        cx509.NameAttribute(
                            cx509.ObjectIdentifier("2.5.4.8"), value="California"
                        )
                    ]
                ),
                cx509.RelativeDistinguishedName(
                    [
                        cx509.NameAttribute(
                            cx509.ObjectIdentifier("2.5.4.7"), value="San Francisco"
                        )
                    ]
                ),
                cx509.RelativeDistinguishedName(
                    [
                        cx509.NameAttribute(
                            cx509.ObjectIdentifier("2.5.4.10"), value="My Company"
                        )
                    ]
                ),
                cx509.RelativeDistinguishedName(
                    [
                        cx509.NameAttribute(
                            cx509.ObjectIdentifier("2.5.4.3"), value="mysite.com"
                        )
                    ]
                ),
            ],
        ),
        (
            ("DNS", "some.invalid_doma.in"),
            salt.exceptions.CommandExecutionError,
            "at position 8.*not allowed$",
        ),
        (
            ("DNS", "some..invalid-doma.in"),
            salt.exceptions.CommandExecutionError,
            "Empty Label",
        ),
        (
            ("DNS", "invalid*.wild.card"),
            salt.exceptions.CommandExecutionError,
            "at position 8.*not allowed",
        ),
        (
            ("DNS", "invalid.*.wild.card"),
            salt.exceptions.CommandExecutionError,
            "at position 1.*not allowed",
        ),
        (
            ("DNS", "*..whats.this"),
            salt.exceptions.CommandExecutionError,
            "Empty label",
        ),
        (
            ("DNS", 42),
            salt.exceptions.SaltInvocationError,
            "Expected string value, got int",
        ),
        (
            ("DNS", ""),
            salt.exceptions.CommandExecutionError,
            "Empty domain",
        ),
        (
            ("DNS", "ἀνεῤῥίφθω.κύβος͵.gr"),
            salt.exceptions.CommandExecutionError,
            "not allowed at position 6 in 'κύβος͵'$",
        ),
        (
            ("DNS", "می\u200cخواهم\u200c.iran"),
            salt.exceptions.CommandExecutionError,
            r"Joiner U\+200C not allowed at position 9 in '.*'",
        ),
        (
            ("DNS", ".*.wildcard-dot.test"),
            salt.exceptions.CommandExecutionError,
            "Wildcards and leading dots cannot be present together",
        ),
        (
            ("email", "invalid@*.mail.address"),
            salt.exceptions.CommandExecutionError,
            "Wildcards are not allowed in this context",
        ),
        (
            ("email", "invalid@.mail.address"),
            salt.exceptions.CommandExecutionError,
            "Leading dots are not allowed in this context",
        ),
        (
            ("email", "Invalid Email <invalid@mail.address>"),
            salt.exceptions.CommandExecutionError,
            "not allowed$",
        ),
        (
            ("IP", "this is not an IP address"),
            salt.exceptions.CommandExecutionError,
            "does not seem to be an IP address or network range.",
        ),
        (
            ("URI", "https://*.χάος.σκάλα.gr"),
            salt.exceptions.CommandExecutionError,
            "Wildcards are not allowed in this context",
        ),
        (
            ("URI", "https://.invalid.host"),
            salt.exceptions.CommandExecutionError,
            "Leading dots are not allowed in this context",
        ),
        (
            ("dirName", "Et tu, Brute?"),
            salt.exceptions.CommandExecutionError,
            "Failed parsing rfc4514 dirName string",
        ),
        (
            ("otherName", "otherName:1.2.3.4;UTF8:some other identifier"),
            salt.exceptions.SaltInvocationError,
            "otherName is currently not implemented",
        ),
        (
            ("invalidType", "L'état c'est moi!"),
            salt.exceptions.CommandExecutionError,
            "GeneralName type invalidtype is invalid",
        ),
    ],
)
def test_parse_general_names(inpt, cls, parsed):
    if issubclass(cls, Exception):
        with pytest.raises(cls, match=parsed):
            x509._parse_general_names([inpt])
        return
    expected = cls(parsed)
    res = x509._parse_general_names([inpt])
    if inpt[0] == "dirName":
        assert res[0].value == expected
    else:
        assert res[0] == expected


@pytest.mark.parametrize(
    "inpt,cls,parsed",
    [
        (("email", "me@example.com"), cx509.RFC822Name, "me@example.com"),
        (
            ("URI", "https://www.example.com"),
            cx509.UniformResourceIdentifier,
            "https://www.example.com",
        ),
        (("DNS", "example.com"), cx509.DNSName, "example.com"),
        (("DNS", "*.example.com"), cx509.DNSName, "*.example.com"),
        (("DNS", ".example.com"), cx509.DNSName, ".example.com"),
        (
            ("DNS", "invalid*.wild.card"),
            salt.exceptions.CommandExecutionError,
            "at position 8.*not allowed",
        ),
        (
            ("DNS", "invalid.*.wild.card"),
            salt.exceptions.CommandExecutionError,
            "at position 1.*not allowed",
        ),
        (
            ("DNS", ".*.wildcard-dot.test"),
            salt.exceptions.CommandExecutionError,
            "Wildcards and leading dots cannot be present together",
        ),
        (
            ("DNS", "gott.würfelt.nicht"),
            salt.exceptions.CommandExecutionError,
            "Cannot encode non-ASCII strings",
        ),
        (
            ("DNS", "some.invalid_doma.in"),
            salt.exceptions.CommandExecutionError,
            "at position 8.*not allowed$",
        ),
        (
            ("DNS", "some..invalid-doma.in"),
            salt.exceptions.CommandExecutionError,
            "Empty Label",
        ),
        (
            ("DNS", 42),
            salt.exceptions.SaltInvocationError,
            "Expected string value, got int",
        ),
        (
            ("DNS", ""),
            salt.exceptions.CommandExecutionError,
            "Empty domain",
        ),
        (
            ("DNS", "*..whats.this"),
            salt.exceptions.CommandExecutionError,
            "Empty label",
        ),
        (
            ("email", "invalid@*.mail.address"),
            salt.exceptions.CommandExecutionError,
            "Wildcards are not allowed in this context",
        ),
        (
            ("email", "invalid@.mail.address"),
            salt.exceptions.CommandExecutionError,
            "Leading dots are not allowed in this context",
        ),
        (
            ("email", "Invalid Email <invalid@mail.address>"),
            salt.exceptions.CommandExecutionError,
            "not allowed$",
        ),
        (
            ("URI", "https://.invalid.host"),
            salt.exceptions.CommandExecutionError,
            "Leading dots are not allowed in this context",
        ),
    ],
)
def test_parse_general_names_without_idna(inpt, cls, parsed):
    with patch("salt.utils.x509.HAS_IDNA", False):
        if issubclass(cls, Exception):
            with pytest.raises(cls, match=parsed):
                x509._parse_general_names([inpt])
            return
        expected = cls(parsed)
        res = x509._parse_general_names([inpt])
        if inpt[0] == "dirName":
            assert res[0].value == expected
        else:
            assert res[0] == expected


@pytest.mark.parametrize(
    "inpt",
    [
        (("RID", "3.2.3.4")),
        (("RID", "1.2.3.4a")),
        (("IP", "13.37.1337.37")),
        (("IP", "13a.37.13.0/24")),
        (("IP", "13.37.13.0:255.255.255.0")),
        (("IP", "20010db8:85a3:0000:0000:8a2e:0370:7334")),
        (("IP", "2001:db8:abcd:0012::0.64")),
        (("dirName", "CC=US,ST=California,L=San Francisco,O=My Company,CN=mysite.com")),
    ],
)
def test_parse_general_names_rejects_invalid(inpt):
    with pytest.raises(salt.exceptions.CommandExecutionError):
        x509._parse_general_names([inpt])


@pytest.mark.parametrize(
    "inpt,expected",
    [
        pytest.param(
            "CN=example.com,O=Example Inc,C=US",
            [
                cx509.NameAttribute(x509.NAME_ATTRS_OID["C"], "US"),
                cx509.NameAttribute(x509.NAME_ATTRS_OID["O"], "Example Inc"),
                cx509.NameAttribute(x509.NAME_ATTRS_OID["CN"], "example.com"),
            ],
        ),
        pytest.param(
            ["C=US", "O=Example Inc", "CN=example.com"],
            [
                cx509.NameAttribute(x509.NAME_ATTRS_OID["C"], "US"),
                cx509.NameAttribute(x509.NAME_ATTRS_OID["O"], "Example Inc"),
                cx509.NameAttribute(x509.NAME_ATTRS_OID["CN"], "example.com"),
            ],
        ),
        pytest.param(
            ["C=US", "O=Example Inc", "OU=foo+CN=example.com"],
            [
                cx509.NameAttribute(x509.NAME_ATTRS_OID["C"], "US"),
                cx509.NameAttribute(x509.NAME_ATTRS_OID["O"], "Example Inc"),
                cx509.RelativeDistinguishedName(
                    {
                        cx509.NameAttribute(x509.NAME_ATTRS_OID["CN"], "example.com"),
                        cx509.NameAttribute(x509.NAME_ATTRS_OID["OU"], "foo"),
                    }
                ),
            ],
        ),
        (
            {"CN": "example.com", "O": "Example Inc", "C": "US", "irrelevant": "bar"},
            [
                cx509.NameAttribute(x509.NAME_ATTRS_OID["C"], "US"),
                cx509.NameAttribute(x509.NAME_ATTRS_OID["O"], "Example Inc"),
                cx509.NameAttribute(x509.NAME_ATTRS_OID["CN"], "example.com"),
            ],
        ),
    ],
)
def test_get_dn(inpt, expected):
    expected_parsed = [
        (
            cx509.RelativeDistinguishedName({x})
            if not isinstance(x, cx509.RelativeDistinguishedName)
            else x
        )
        for x in expected
    ]
    res = x509._get_dn(inpt)
    assert res.rdns == expected_parsed


@pytest.mark.parametrize(
    "inpt,expected",
    [
        (
            cx509.Extension(
                cx509.BasicConstraints.oid,
                value=cx509.BasicConstraints(ca=True, path_length=2),
                critical=True,
            ),
            {"ca": True, "critical": True, "pathlen": 2},
        ),
        (
            cx509.Extension(
                cx509.KeyUsage.oid,
                value=cx509.KeyUsage(
                    digital_signature=True,
                    content_commitment=False,
                    key_encipherment=False,
                    data_encipherment=False,
                    key_agreement=False,
                    key_cert_sign=False,
                    crl_sign=False,
                    encipher_only=False,
                    decipher_only=False,
                ),
                critical=True,
            ),
            {
                "cRLSign": False,
                "critical": True,
                "dataEncipherment": False,
                "decipherOnly": False,
                "digitalSignature": True,
                "encipherOnly": False,
                "keyAgreement": False,
                "keyCertSign": False,
                "keyEncipherment": False,
                "nonRepudiation": False,
            },
        ),
        (
            cx509.Extension(
                cx509.ExtendedKeyUsage.oid,
                value=cx509.ExtendedKeyUsage(
                    [
                        x509.EXTENDED_KEY_USAGE_OID["OCSPSigning"],
                        x509.EXTENDED_KEY_USAGE_OID["msCodeCom"],
                        cx509.ObjectIdentifier("2.3.4.55"),
                    ]
                ),
                critical=True,
            ),
            {
                "critical": True,
                "value": ["OCSPSigning", "1.3.6.1.4.1.311.2.1.22", "2.3.4.55"],
            },
        ),
        (
            cx509.Extension(
                cx509.SubjectKeyIdentifier.oid,
                value=cx509.SubjectKeyIdentifier(
                    b"\xbb\xaf~\x02=\xfa\xa6\xf1<\x84\x8e\xad\xee8\x98\xec\xd922\xd4"
                ),
                critical=False,
            ),
            {
                "critical": False,
                "value": "BB:AF:7E:02:3D:FA:A6:F1:3C:84:8E:AD:EE:38:98:EC:D9:32:32:D4",
            },
        ),
        (
            cx509.Extension(
                cx509.AuthorityKeyIdentifier.oid,
                value=cx509.AuthorityKeyIdentifier(
                    key_identifier=b"\xde\xad\xbe\xef",
                    authority_cert_issuer=[
                        cx509.DirectoryName(
                            cx509.Name(
                                [
                                    cx509.RelativeDistinguishedName(
                                        [
                                            cx509.NameAttribute(
                                                cx509.ObjectIdentifier("2.5.4.3"),
                                                value="foo",
                                            )
                                        ]
                                    )
                                ]
                            )
                        )
                    ],
                    authority_cert_serial_number=1337,
                ),
                critical=False,
            ),
            {
                "critical": False,
                "issuer": ["dirName:CN=foo"],
                "issuer_sn": "05:39",
                "keyid": "DE:AD:BE:EF",
            },
        ),
        (
            cx509.Extension(
                cx509.IssuerAlternativeName.oid,
                value=cx509.IssuerAlternativeName(
                    [
                        cx509.RFC822Name("ca@example.com"),
                        cx509.DNSName("example.com"),
                        cx509.DNSName("example.io"),
                    ]
                ),
                critical=False,
            ),
            {
                "critical": False,
                "value": ["mail:ca@example.com", "DNS:example.com", "DNS:example.io"],
            },
        ),
        (
            cx509.Extension(
                cx509.CertificateIssuer.oid,
                value=cx509.CertificateIssuer(
                    [
                        cx509.RFC822Name("ca@example.com"),
                        cx509.DNSName("example.com"),
                        cx509.DNSName("example.io"),
                    ]
                ),
                critical=False,
            ),
            {
                "critical": False,
                "value": ["mail:ca@example.com", "DNS:example.com", "DNS:example.io"],
            },
        ),
        (
            cx509.Extension(
                cx509.AuthorityInformationAccess.oid,
                value=cx509.AuthorityInformationAccess(
                    [
                        cx509.AccessDescription(
                            access_method=x509.ACCESS_OID["OCSP"],
                            access_location=cx509.UniformResourceIdentifier(
                                "http://ocsp.example.com/"
                            ),
                        ),
                        cx509.AccessDescription(
                            access_method=x509.ACCESS_OID["OCSP"],
                            access_location=cx509.UniformResourceIdentifier(
                                "http://ocsp2.example.com/"
                            ),
                        ),
                    ]
                ),
                critical=False,
            ),
            {
                "critical": False,
                "value": [
                    {"OCSP": "http://ocsp.example.com/"},
                    {"OCSP": "http://ocsp2.example.com/"},
                ],
            },
        ),
        (
            cx509.Extension(
                cx509.SubjectAlternativeName.oid,
                value=cx509.SubjectAlternativeName(
                    [cx509.DNSName("example.io"), cx509.RFC822Name("hello@example.io")]
                ),
                critical=False,
            ),
            {"critical": False, "value": ["DNS:example.io", "mail:hello@example.io"]},
        ),
        (
            cx509.Extension(
                cx509.CRLDistributionPoints.oid,
                value=cx509.CRLDistributionPoints(
                    [
                        cx509.DistributionPoint(
                            full_name=[
                                cx509.UniformResourceIdentifier(
                                    "http://example.com/myca.crl"
                                )
                            ],
                            relative_name=None,
                            reasons=frozenset([cx509.ReasonFlags("keyCompromise")]),
                            crl_issuer=[cx509.DNSName("example.org")],
                        ),
                        cx509.DistributionPoint(
                            full_name=None,
                            relative_name=cx509.RelativeDistinguishedName(
                                [
                                    cx509.NameAttribute(
                                        cx509.ObjectIdentifier("2.5.4.11"), value="foo"
                                    ),
                                    cx509.NameAttribute(
                                        cx509.ObjectIdentifier("2.5.4.3"), value="Smith"
                                    ),
                                ]
                            ),
                            reasons=None,
                            crl_issuer=[cx509.DNSName("example.org")],
                        ),
                    ]
                ),
                critical=False,
            ),
            {
                "critical": False,
                "value": [
                    {
                        "crlissuer": ["DNS:example.org"],
                        "fullname": ["URI:http://example.com/myca.crl"],
                        "reasons": ["keyCompromise"],
                        "relativename": None,
                    },
                    {
                        "crlissuer": ["DNS:example.org"],
                        "fullname": [],
                        "reasons": [],
                        "relativename": "OU=foo+CN=Smith",
                    },
                ],
            },
        ),
        (
            cx509.Extension(
                cx509.FreshestCRL.oid,
                value=cx509.FreshestCRL(
                    [
                        cx509.DistributionPoint(
                            full_name=[
                                cx509.UniformResourceIdentifier(
                                    "http://example.com/myca.crl"
                                )
                            ],
                            relative_name=None,
                            reasons=frozenset([cx509.ReasonFlags("keyCompromise")]),
                            crl_issuer=[cx509.DNSName("example.org")],
                        ),
                        cx509.DistributionPoint(
                            full_name=None,
                            relative_name=cx509.RelativeDistinguishedName(
                                [
                                    cx509.NameAttribute(
                                        cx509.ObjectIdentifier("2.5.4.11"), value="foo"
                                    ),
                                    cx509.NameAttribute(
                                        cx509.ObjectIdentifier("2.5.4.3"), value="Smith"
                                    ),
                                ]
                            ),
                            reasons=None,
                            crl_issuer=[cx509.DNSName("example.org")],
                        ),
                    ]
                ),
                critical=False,
            ),
            {
                "critical": False,
                "value": [
                    {
                        "crlissuer": ["DNS:example.org"],
                        "fullname": ["URI:http://example.com/myca.crl"],
                        "reasons": ["keyCompromise"],
                        "relativename": None,
                    },
                    {
                        "crlissuer": ["DNS:example.org"],
                        "fullname": [],
                        "reasons": [],
                        "relativename": "OU=foo+CN=Smith",
                    },
                ],
            },
        ),
        (
            cx509.Extension(
                cx509.IssuingDistributionPoint.oid,
                value=cx509.IssuingDistributionPoint(
                    full_name=[
                        cx509.UniformResourceIdentifier("http://example.com/myca.crl")
                    ],
                    relative_name=None,
                    only_some_reasons=frozenset([cx509.ReasonFlags("keyCompromise")]),
                    only_contains_user_certs=True,
                    only_contains_ca_certs=False,
                    only_contains_attribute_certs=False,
                    indirect_crl=False,
                ),
                critical=False,
            ),
            {
                "critical": False,
                "fullname": ["URI:http://example.com/myca.crl"],
                "indirectCRL": False,
                "onlyAA": False,
                "onlyCA": False,
                "onlyuser": True,
                "onysomereasons": ["keyCompromise"],
                "relativename": None,
            },
        ),
        (
            cx509.Extension(
                cx509.CertificatePolicies.oid,
                value=cx509.CertificatePolicies(
                    [
                        cx509.PolicyInformation(
                            policy_identifier=cx509.ObjectIdentifier("1.2.3.4.5"),
                            policy_qualifiers=[
                                "https://my.ca.com/pratice_statement",
                                cx509.UserNotice(
                                    notice_reference=cx509.NoticeReference(
                                        organization="myorg", notice_numbers=[1, 2, 3]
                                    ),
                                    explicit_text="mytext",
                                ),
                            ],
                        )
                    ]
                ),
                critical=False,
            ),
            {
                "critical": False,
                "value": [
                    {
                        "1.2.3.4.5": [
                            {
                                "practice_statement": "https://my.ca.com/pratice_statement"
                            },
                            {
                                "explicit_text": "mytext",
                                "notice_numbers": [1, 2, 3],
                                "organizataion": "myorg",
                            },
                        ]
                    }
                ],
            },
        ),
        (
            cx509.Extension(
                cx509.PolicyConstraints.oid,
                value=cx509.PolicyConstraints(
                    require_explicit_policy=4, inhibit_policy_mapping=2
                ),
                critical=False,
            ),
            {"critical": False, "inhibitPolicyMapping": 2, "requireExplicitPolicy": 4},
        ),
        (
            cx509.Extension(
                cx509.InhibitAnyPolicy.oid,
                value=cx509.InhibitAnyPolicy(1),
                critical=False,
            ),
            {"critical": False, "value": 1},
        ),
        (
            cx509.Extension(
                cx509.NameConstraints.oid,
                value=cx509.NameConstraints(
                    [
                        cx509.IPAddress(ipaddress.ip_network("192.168.0.0/16")),
                        cx509.RFC822Name(".example.com"),
                    ],
                    [cx509.RFC822Name(".com")],
                ),
                critical=False,
            ),
            {
                "critical": False,
                "excluded": ["mail:.com"],
                "permitted": ["IP:192.168.0.0/16", "mail:.example.com"],
            },
        ),
        (
            cx509.Extension(
                cx509.OCSPNoCheck.oid, value=cx509.OCSPNoCheck(), critical=False
            ),
            {"critical": False, "value": True},
        ),
        (
            cx509.Extension(
                cx509.TLSFeature.oid,
                value=cx509.TLSFeature([cx509.TLSFeatureType.status_request]),
                critical=False,
            ),
            {"critical": False, "value": ["status_request"]},
        ),
        (
            cx509.Extension(
                cx509.CRLNumber.oid, value=cx509.CRLNumber(3), critical=False
            ),
            {"critical": False, "value": 3},
        ),
        (
            cx509.Extension(
                cx509.DeltaCRLIndicator.oid,
                value=cx509.DeltaCRLIndicator(3),
                critical=False,
            ),
            {"critical": False, "value": 3},
        ),
        (
            cx509.Extension(
                cx509.CRLReason.oid,
                value=cx509.CRLReason(cx509.ReasonFlags("superseded")),
                critical=False,
            ),
            {"critical": False, "value": "superseded"},
        ),
        (
            cx509.Extension(
                cx509.InvalidityDate.oid,
                value=cx509.InvalidityDate(
                    datetime.strptime("2022-10-11 13:37:42", "%Y-%m-%d %H:%M:%S")
                ),
                critical=False,
            ),
            {"critical": False, "value": "2022-10-11 13:37:42"},
        ),
    ],
)
def test_render_extension(inpt, expected):
    ret = x509.render_extension(inpt)
    assert ret == expected


@pytest.fixture
def ca_cert():
    return """\
-----BEGIN CERTIFICATE-----
MIIDODCCAiCgAwIBAgIIbfpgqP0VGPgwDQYJKoZIhvcNAQELBQAwKzELMAkGA1UE
BhMCVVMxDTALBgNVBAMMBFRlc3QxDTALBgNVBAoMBFNhbHQwHhcNMjIxMTE1MTQw
NDMzWhcNMzIxMTEyMTQwNDMzWjArMQswCQYDVQQGEwJVUzENMAsGA1UEAwwEVGVz
dDENMAsGA1UECgwEU2FsdDCCASIwDQYJKoZIhvcNAQEBBQADggEPADCCAQoCggEB
AOGTScvrjcEt6vsJcG9RUp6fKaDNDWZnJET0omanK9ZwaoGpJPp8UDYe/8ADeI7N
10wdyB4oDM9gRDjInBtdQO/PsrmKZF6LzqVFgLMxu2up+PHMi9z6B2P4esIAzMu9
PYxc9zH4HzLImHqscVD2HCabsjp9X134Af7hVY5NN/W/4qTP7uOM20wSG2TPI6+B
tA9VyPbEPMPRzXzrqc45rVYe6kb2bT84GE93Vcu/e5JZ/k2AKD8Hoa2cxLPsTLq5
igl+D+k+dfUtiABiKPvVQiYBsD1fyHDn2m7B6pCgvrGqHjsoAKufgFnXy6PJRg7n
vQfaxSiusM5s+VS+fjlvgwsCAwEAAaNgMF4wDwYDVR0TBAgwBgEB/wIBATALBgNV
HQ8EBAMCAQYwHQYDVR0OBBYEFFzy8fRTKSOe7kBakqO0Ki71potnMB8GA1UdIwQY
MBaAFFzy8fRTKSOe7kBakqO0Ki71potnMA0GCSqGSIb3DQEBCwUAA4IBAQBZS4MP
fXYPoGZ66seM+0eikScZHirbRe8vHxHkujnTBUjQITKm86WeQgeBCD2pobgBGZtt
5YFozM4cERqY7/1BdemUxFvPmMFFznt0TM5w+DfGWVK8un6SYwHnmBbnkWgX4Srm
GsL0HHWxVXkGnFGFk6Sbo3vnN7CpkpQTWFqeQQ5rHOw91pt7KnNZwc6I3ZjrCUHJ
+UmKKrga16a4Q+8FBpYdphQU609npo/0zuaE6FyiJYlW3tG+mlbbNgzY/+eUaxt2
9Bp9mtA+Hkox551Mfpq45Oi+ehwMt0xjZCjuFCM78oiUdHCGO+EmcT7ogiYALiOF
LN1w5sybsYwIw6QN
-----END CERTIFICATE-----
"""


@pytest.fixture
def ca_key():
    return """\
-----BEGIN RSA PRIVATE KEY-----
MIIEowIBAAKCAQEA4ZNJy+uNwS3q+wlwb1FSnp8poM0NZmckRPSiZqcr1nBqgakk
+nxQNh7/wAN4js3XTB3IHigMz2BEOMicG11A78+yuYpkXovOpUWAszG7a6n48cyL
3PoHY/h6wgDMy709jFz3MfgfMsiYeqxxUPYcJpuyOn1fXfgB/uFVjk039b/ipM/u
44zbTBIbZM8jr4G0D1XI9sQ8w9HNfOupzjmtVh7qRvZtPzgYT3dVy797kln+TYAo
PwehrZzEs+xMurmKCX4P6T519S2IAGIo+9VCJgGwPV/IcOfabsHqkKC+saoeOygA
q5+AWdfLo8lGDue9B9rFKK6wzmz5VL5+OW+DCwIDAQABAoIBAFfImc9hu6iR1gAb
jEXFwAE6r1iEc9KGEPdEvG52X/jzhn8u89UGy7BEIAL5VtE8Caz1agtSSqnpLKNs
blO31q18hnDuCmFAxwpKIeuaTvV3EAoJL+Su6HFfIWaeKRSgcHNPOmOXy4xXw/75
XJ/FJu9fZ9ybLaHEAgLObh0Sr9RSPQbZ72ZawPP8+5WCbR+2w90RApHXQL0piSbW
lIx1NE6o5wQb3vik8z/k5FqLCY2a8++WNyfvS+WWFY5WXGI7ZiDDQk46gnslquH2
Lon5CEn3JlTGQFhxaaa2ivssscf2lA2Rvm2E8o1rdZJS2OpSE0ai4TXY9XnyjZj1
5usWIwECgYEA+3Mwu03A7PyLEBksS/u3MSo/176S9lF/uXcecQNdhAIalUZ8AgV3
7HP2yI9ZC0ekA809ZzFjGFostXm9VfUOEZ549jLOMzvBtCdaI0aBUE8icu52fX4r
fT2NY6hYgz5/fxD8sq1XH/fqNNexABwtViH6YAly/9A1/8M3BOWt72UCgYEA5ag8
sIfiBUoWd1sS6qHDuugWlpx4ZWYC/59XEJyCN2wioP8qFji/aNZxF1wLfyQe/zaa
YBFusjsBnSfBU1p4UKCRHWQ9/CnC0DzqTkyKC4Fv8GuxgywNm5W9gPKk7idHP7mw
e+7Uvf1pOQccqEPh7yltpW+Xw27gfsC2DMAIGa8CgYByv/q5P56PiCCeVB6W/mR3
l2RTPLEsn7y+EtJdmL+QgrVG8kedVImJ6tHwbRqhvyvmYD9pXGxwrJZCqy/wjkjB
WaSyFjVrxBV99Yd5Ga/hyntaH+ELHA0UtoZTuHvMSTU9866ei+R6vlSvkM9B0ZoO
+KqeMTG99HLwKVJudbKO0QKBgQCd33U49XBOqoufKSBr4yAmUH2Ws6GgMuxExUiY
xr5NUyzK+B36gLA0ZZYAtOnCURZt4x9kgxdRtnZ5jma74ilrY7XeOpbRzfN6KyX3
BW6wUh6da6rvvUztc5Z+Gk9+18mG6SOFTr04jgfTiCwPD/s06YnSfFAbrRDukZOU
WD45SQKBgBvjSwl3AbPoJnRjZjGuCUMKQKrLm30xCeorxasu+di/4YV5Yd8VUjaO
mYyqXW6bQndKLuXT+AXtCd/Xt2sI96z8mc0G5fImDUxQjMUuS3RyQK357cEOu8Zy
HdI7Pfaf/l0HozAw/Al+LXbpmSBdfmz0U/EGAKRqXMW5+vQ7XHXD
-----END RSA PRIVATE KEY-----"""


def test_build_crl_accounts_for_local_time_zone(ca_key, ca_cert):
    curr_time = datetime.now(tz=timezone(timedelta(hours=1)))
    curr_time_naive = curr_time.replace(tzinfo=None)

    def dtn(tz=None):
        if tz is None:
            return curr_time_naive
        return curr_time

    curr_time_utc = curr_time.astimezone(timezone.utc).replace(microsecond=0)
    curr_time_utc_naive = curr_time_utc.replace(tzinfo=None)
    privkey = cprim.serialization.load_pem_private_key(ca_key.encode(), password=None)
    cert = cx509.load_pem_x509_certificate(ca_cert.encode())
    with patch("salt.utils.x509.datetime") as fakedate:
        fakedate.today.return_value = curr_time_naive
        fakedate.now.side_effect = dtn
        fakedate.utcnow.return_value = curr_time_utc_naive
        builder, _ = x509.build_crl(privkey, [], signing_cert=cert)
        crl = builder.sign(privkey, algorithm=cprim.hashes.SHA256())
    try:
        assert crl.last_update_utc == curr_time_utc
    except AttributeError:
        assert crl.last_update == curr_time_utc_naive



"""
    :synopsis: Base class for kernelpkg modules
    :platform: Linux
    :maturity: develop
    .. versionadded:: 2018.3.0
"""

# pylint: disable=invalid-name,no-member


# Salt testing libs
try:
    from salt.exceptions import CommandExecutionError
    from tests.support.mock import MagicMock, patch
except ImportError:
    pass


class KernelPkgTestCase:
    """
    Test cases shared by all kernelpkg virtual modules
    """

    def test_active(self):
        """
        Test - Return return the active kernel version
        """
        self.assertEqual(self._kernelpkg.active(), self.KERNEL_LIST[0])

    def test_latest_available_no_results(self):
        """
        Test - Return the latest available kernel version
        """
        mock = MagicMock(return_value="")
        with patch.dict(self._kernelpkg.__salt__, {"pkg.latest_version": mock}):
            with patch.object(
                self._kernelpkg, "active", return_value=self.KERNEL_LIST[0]
            ):
                self.assertEqual(
                    self._kernelpkg.latest_available(), self.KERNEL_LIST[-1]
                )

    def test_latest_available_at_latest(self):
        """
        Test - Return the latest available kernel version
        """
        mock = MagicMock(return_value=self.LATEST)
        with patch.dict(self._kernelpkg.__salt__, {"pkg.latest_version": mock}):
            with patch.object(
                self._kernelpkg, "active", return_value=self.KERNEL_LIST[-1]
            ):
                self.assertEqual(
                    self._kernelpkg.latest_available(), self.KERNEL_LIST[-1]
                )

    def test_latest_available_with_updates(self):
        """
        Test - Return the latest available kernel version
        """
        mock = MagicMock(return_value=self.LATEST)
        with patch.dict(self._kernelpkg.__salt__, {"pkg.latest_version": mock}):
            with patch.object(
                self._kernelpkg, "active", return_value=self.KERNEL_LIST[0]
            ):
                self.assertEqual(
                    self._kernelpkg.latest_available(), self.KERNEL_LIST[-1]
                )

    def test_latest_installed_with_updates(self):
        """
        Test - Return the latest installed kernel version
        """
        with patch.object(self._kernelpkg, "active", return_value=self.KERNEL_LIST[0]):
            with patch.object(
                self._kernelpkg, "list_installed", return_value=self.KERNEL_LIST
            ):
                self.assertEqual(
                    self._kernelpkg.latest_installed(), self.KERNEL_LIST[-1]
                )

    def test_latest_installed_at_latest(self):
        """
        Test - Return the latest installed kernel version
        """
        with patch.object(self._kernelpkg, "active", return_value=self.KERNEL_LIST[-1]):
            with patch.object(
                self._kernelpkg, "list_installed", return_value=self.KERNEL_LIST
            ):
                self.assertEqual(
                    self._kernelpkg.latest_installed(), self.KERNEL_LIST[-1]
                )

    def test_needs_reboot_with_update(self):
        """
        Test - Return True if a new kernel is ready to be booted
        """
        with patch.object(self._kernelpkg, "active", return_value=self.KERNEL_LIST[0]):
            with patch.object(
                self._kernelpkg, "latest_installed", return_value=self.KERNEL_LIST[1]
            ):
                self.assertTrue(self._kernelpkg.needs_reboot())

    def test_needs_reboot_at_latest(self):
        """
        Test - Return True if a new kernel is ready to be booted
        """
        with patch.object(self._kernelpkg, "active", return_value=self.KERNEL_LIST[1]):
            with patch.object(
                self._kernelpkg, "latest_installed", return_value=self.KERNEL_LIST[1]
            ):
                self.assertFalse(self._kernelpkg.needs_reboot())

    def test_needs_reboot_order_inverted(self):
        """
        Test - Return True if a new kernel is ready to be booted
        """
        with patch.object(self._kernelpkg, "active", return_value=self.KERNEL_LIST[1]):
            with patch.object(
                self._kernelpkg, "latest_installed", return_value=self.KERNEL_LIST[0]
            ):
                self.assertFalse(self._kernelpkg.needs_reboot())

    def test_upgrade_not_needed_with_reboot(self):
        """
        Test - Upgrade function when no upgrade is available and reboot has been requested
        """
        with patch.object(self._kernelpkg, "active", return_value=self.KERNEL_LIST[-1]):
            with patch.object(
                self._kernelpkg, "list_installed", return_value=self.KERNEL_LIST
            ):
                result = self._kernelpkg.upgrade(reboot=True)
                self.assertIn("upgrades", result)
                self.assertEqual(result["active"], self.KERNEL_LIST[-1])
                self.assertEqual(result["latest_installed"], self.KERNEL_LIST[-1])
                self.assertEqual(result["reboot_requested"], True)
                self.assertEqual(result["reboot_required"], False)
                self._kernelpkg.__salt__["system.reboot"].assert_not_called()

    def test_upgrade_not_needed_without_reboot(self):
        """
        Test - Upgrade function when no upgrade is available and no reboot has been requested
        """
        with patch.object(self._kernelpkg, "active", return_value=self.KERNEL_LIST[-1]):
            with patch.object(
                self._kernelpkg, "list_installed", return_value=self.KERNEL_LIST
            ):
                result = self._kernelpkg.upgrade(reboot=False)
                self.assertIn("upgrades", result)
                self.assertEqual(result["active"], self.KERNEL_LIST[-1])
                self.assertEqual(result["latest_installed"], self.KERNEL_LIST[-1])
                self.assertEqual(result["reboot_requested"], False)
                self.assertEqual(result["reboot_required"], False)
                self._kernelpkg.__salt__["system.reboot"].assert_not_called()

    def test_upgrade_needed_with_reboot(self):
        """
        Test - Upgrade function when an upgrade is available and reboot has been requested
        """
        with patch.object(self._kernelpkg, "active", return_value=self.KERNEL_LIST[0]):
            with patch.object(
                self._kernelpkg, "list_installed", return_value=self.KERNEL_LIST
            ):
                result = self._kernelpkg.upgrade(reboot=True)
                self.assertIn("upgrades", result)
                self.assertEqual(result["active"], self.KERNEL_LIST[0])
                self.assertEqual(result["latest_installed"], self.KERNEL_LIST[-1])
                self.assertEqual(result["reboot_requested"], True)
                self.assertEqual(result["reboot_required"], True)
                self._kernelpkg.__salt__["system.reboot"].assert_called_once()

    def test_upgrade_needed_without_reboot(self):
        """
        Test - Upgrade function when an upgrade is available and no reboot has been requested
        """
        with patch.object(self._kernelpkg, "active", return_value=self.KERNEL_LIST[0]):
            with patch.object(
                self._kernelpkg, "list_installed", return_value=self.KERNEL_LIST
            ):
                result = self._kernelpkg.upgrade(reboot=False)
                self.assertIn("upgrades", result)
                self.assertEqual(result["active"], self.KERNEL_LIST[0])
                self.assertEqual(result["latest_installed"], self.KERNEL_LIST[-1])
                self.assertEqual(result["reboot_requested"], False)
                self.assertEqual(result["reboot_required"], True)
                self._kernelpkg.__salt__["system.reboot"].assert_not_called()

    def test_upgrade_available_true(self):
        """
        Test - upgrade_available
        """
        with patch.object(
            self._kernelpkg, "latest_available", return_value=self.KERNEL_LIST[-1]
        ):
            with patch.object(
                self._kernelpkg, "latest_installed", return_value=self.KERNEL_LIST[0]
            ):
                self.assertTrue(self._kernelpkg.upgrade_available())

    def test_upgrade_available_false(self):
        """
        Test - upgrade_available
        """
        with patch.object(
            self._kernelpkg, "latest_available", return_value=self.KERNEL_LIST[-1]
        ):
            with patch.object(
                self._kernelpkg, "latest_installed", return_value=self.KERNEL_LIST[-1]
            ):
                self.assertFalse(self._kernelpkg.upgrade_available())

    def test_upgrade_available_inverted(self):
        """
        Test - upgrade_available
        """
        with patch.object(
            self._kernelpkg, "latest_available", return_value=self.KERNEL_LIST[0]
        ):
            with patch.object(
                self._kernelpkg, "latest_installed", return_value=self.KERNEL_LIST[-1]
            ):
                self.assertFalse(self._kernelpkg.upgrade_available())

    def test_remove_active(self):
        """
        Test - remove kernel package
        """
        mock = MagicMock(return_value={"retcode": 0, "stderr": []})
        with patch.dict(self._kernelpkg.__salt__, {"cmd.run_all": mock}):
            with patch.object(
                self._kernelpkg, "active", return_value=self.KERNEL_LIST[-1]
            ):
                with patch.object(
                    self._kernelpkg, "list_installed", return_value=self.KERNEL_LIST
                ):
                    self.assertRaises(
                        CommandExecutionError,
                        self._kernelpkg.remove,
                        release=self.KERNEL_LIST[-1],
                    )
                    self._kernelpkg.__salt__["cmd.run_all"].assert_not_called()

    def test_remove_invalid(self):
        """
        Test - remove kernel package
        """
        mock = MagicMock(return_value={"retcode": 0, "stderr": []})
        with patch.dict(self._kernelpkg.__salt__, {"cmd.run_all": mock}):
            with patch.object(
                self._kernelpkg, "active", return_value=self.KERNEL_LIST[-1]
            ):
                with patch.object(
                    self._kernelpkg, "list_installed", return_value=self.KERNEL_LIST
                ):
                    self.assertRaises(
                        CommandExecutionError, self._kernelpkg.remove, release="invalid"
                    )
                    self._kernelpkg.__salt__["cmd.run_all"].assert_not_called()



import random

import pytest

import salt.utils.platform

pytestmark = [
    pytest.mark.slow_test,
]


@pytest.fixture(scope="module")
def swarm_timeout():
    timeout = 120
    if salt.utils.platform.spawning_platform():
        timeout *= 2
    return timeout


def test_ping(minion_swarm, salt_cli, swarm_timeout):
    ret = salt_cli.run("test.ping", minion_tgt="*", _timeout=swarm_timeout)
    assert ret.returncode == 0
    assert ret.data
    for minion in minion_swarm:
        assert minion.id in ret.data
        minion_ret = ret.data[minion.id]
        # Sometimes the command times out but doesn't fail, so we catch it
        if isinstance(minion_ret, str) and "Minion did not return" in minion_ret:
            continue
        assert ret.data[minion.id] is True


def test_ping_one(minion_swarm, salt_cli, swarm_timeout):
    minion = random.choice(minion_swarm)
    ret = salt_cli.run("test.ping", minion_tgt=minion.id, _timeout=swarm_timeout)
    assert ret.returncode == 0
    assert ret.data is True



import logging

import pytest

from salt.cloud.clouds import vultrpy as vultr
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {
        vultr: {
            "__utils__": {
                "cloud.fire_event": MagicMock(),
                "cloud.filter_event": MagicMock(),
                "cloud.wait_for_fun": MagicMock(),
                "cloud.bootstrap": MagicMock(),
            },
            "__opts__": {
                "providers": {
                    "vultr01": {
                        "vultr": {
                            "api_key": "super_secret_key",
                            "driver": "vultr",
                        },
                    },
                },
                "sock_dir": "/tmp/sock_dir",
                "transport": "tcp",
            },
            "__active_provider_name__": "my_vultr:vultr",
        }
    }


def test_show_keypair_no_keyname(caplog):
    """
    test salt.cloud.clouds.vultr.show_keypair
    when keyname is not in kwargs
    """
    with caplog.at_level(logging.INFO):
        assert not vultr.show_keypair({})
        assert "A keyname is required." in caplog.text


def test_show_keypair():
    """
    test salt.cloud.clouds.vultr.show_keypair
    when keyname provided
    """
    with patch(
        "salt.cloud.clouds.vultrpy._query", return_value={"test": {"SSHKEYID": "keyID"}}
    ):
        kwargs = {"keyname": "test"}
        assert vultr.show_keypair(kwargs) == {"SSHKEYID": "keyID"}


def test_create_firewall_ssh():
    """
    Test create when setting firewall_group_id and
    ssh_key_names.
    """
    kwargs = {
        "provider": "vultr",
        "enable_private_network": True,
        "ssh_key_names": "key1,key2,key3",
        "startup_script_id": "test_id",
        "firewall_group_id": "f_id",
        "image": 223,
        "size": 13,
        "location": 1,
        "name": "test-vm",
    }
    patch_scripts = patch(
        "salt.cloud.clouds.vultrpy.avail_scripts",
        MagicMock(return_value=["test_id"]),
    )

    patch_firewall = patch(
        "salt.cloud.clouds.vultrpy.avail_firewall_groups",
        MagicMock(return_value=["f_id"]),
    )

    patch_keys = patch(
        "salt.cloud.clouds.vultrpy.avail_keys",
        MagicMock(return_value=["key3", "key2", "key1"]),
    )

    patch_vultrid = patch(
        "salt.cloud.clouds.vultrpy._lookup_vultrid",
        MagicMock(return_value="test_id"),
    )

    mock_query = MagicMock(return_value={"status": 200})
    patch_query = patch("salt.cloud.clouds.vultrpy._query", mock_query)

    patch_show = patch("salt.cloud.clouds.vultrpy.show_instance", MagicMock())

    with (
        patch_scripts
    ), patch_firewall, patch_keys, patch_vultrid, patch_query, patch_show:
        vultr.create(kwargs)
        query_ret = mock_query.call_args.kwargs["data"]
        assert "SSHKEYID=key1%2Ckey2%2Ckey3" in query_ret
        assert "FIREWALLGROUPID=f_id" in query_ret


def test_create_firewall_doesnotexist(caplog):
    """
    Test create when setting firewall_group_id to a firewall
    that does not exist
    """
    kwargs = {
        "provider": "vultr",
        "enable_private_network": True,
        "startup_script_id": "test_id",
        "firewall_group_id": "doesnotexist",
        "image": 223,
        "size": 13,
        "location": 1,
        "name": "test-vm",
    }
    patch_scripts = patch(
        "salt.cloud.clouds.vultrpy.avail_scripts",
        MagicMock(return_value=["test_id"]),
    )

    patch_firewall = patch(
        "salt.cloud.clouds.vultrpy.avail_firewall_groups",
        MagicMock(return_value=["f_id"]),
    )

    patch_keys = patch(
        "salt.cloud.clouds.vultrpy.avail_keys",
        MagicMock(return_value=["key3", "key2", "key1"]),
    )

    patch_vultrid = patch(
        "salt.cloud.clouds.vultrpy._lookup_vultrid",
        MagicMock(return_value="test_id"),
    )

    mock_query = MagicMock(return_value={"status": 200})
    patch_query = patch("salt.cloud.clouds.vultrpy._query", mock_query)

    patch_show = patch("salt.cloud.clouds.vultrpy.show_instance", MagicMock())

    with (
        patch_scripts
    ), patch_firewall, patch_keys, patch_vultrid, patch_query, patch_show:
        with caplog.at_level(logging.INFO):
            ret = vultr.create(kwargs)
            assert (
                "Your Vultr account does not have a firewall group with ID doesnotexist"
                in caplog.text
            )
            assert ret is False


def test_create_ssh_key_ids_doesnotexist(caplog):
    """
    Test create when setting ssh_key_ids that do not
    exist
    """
    kwargs = {
        "provider": "vultr",
        "enable_private_network": True,
        "startup_script_id": "test_id",
        "ssh_key_names": "doesnotexist",
        "image": 223,
        "size": 13,
        "location": 1,
        "name": "test-vm",
    }
    patch_scripts = patch(
        "salt.cloud.clouds.vultrpy.avail_scripts",
        MagicMock(return_value=["test_id"]),
    )

    patch_firewall = patch(
        "salt.cloud.clouds.vultrpy.avail_firewall_groups",
        MagicMock(return_value=["f_id"]),
    )

    patch_keys = patch(
        "salt.cloud.clouds.vultrpy.avail_keys",
        MagicMock(return_value=["key3", "key2", "key1"]),
    )

    patch_vultrid = patch(
        "salt.cloud.clouds.vultrpy._lookup_vultrid",
        MagicMock(return_value="test_id"),
    )

    mock_query = MagicMock(return_value={"status": 200})
    patch_query = patch("salt.cloud.clouds.vultrpy._query", mock_query)

    patch_show = patch("salt.cloud.clouds.vultrpy.show_instance", MagicMock())

    with (
        patch_scripts
    ), patch_firewall, patch_keys, patch_vultrid, patch_query, patch_show:
        with caplog.at_level(logging.INFO):
            ret = vultr.create(kwargs)
            assert (
                "Your Vultr account does not have a key with ID doesnotexist"
                in caplog.text
            )
            assert ret is False



"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>

    Test cases for salt.modules.rsyn
"""

import pytest

import salt.modules.rsync as rsync
from salt.exceptions import CommandExecutionError, SaltInvocationError
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {rsync: {}}


def test_rsync():
    """
    Test for rsync files from src to dst
    """
    with patch.dict(rsync.__salt__, {"config.option": MagicMock(return_value=False)}):
        pytest.raises(SaltInvocationError, rsync.rsync, "", "")

    with patch.dict(
        rsync.__salt__,
        {
            "config.option": MagicMock(return_value="A"),
            "cmd.run_all": MagicMock(side_effect=[OSError(1, "f"), "A"]),
        },
    ):
        with patch.object(rsync, "_check", return_value=["A"]):
            pytest.raises(CommandExecutionError, rsync.rsync, "a", "b")

            assert rsync.rsync("src", "dst") == "A"


def test_version():
    """
    Test for return rsync version
    """
    mock = MagicMock(side_effect=[OSError(1, "f"), "A B C\n"])
    with patch.dict(rsync.__salt__, {"cmd.run_stdout": mock}):
        pytest.raises(CommandExecutionError, rsync.version)

        assert rsync.version() == "C"


def test_rsync_excludes_list():
    """
    Test for rsync files from src to dst with a list of excludes
    """
    mock = {
        "config.option": MagicMock(return_value=False),
        "cmd.run_all": MagicMock(),
    }
    with patch.dict(rsync.__salt__, mock):
        rsync.rsync("src", "dst", exclude=["test/one", "test/two"])
    mock["cmd.run_all"].assert_called_once_with(
        [
            "rsync",
            "-avz",
            "--exclude",
            "test/one",
            "--exclude",
            "test/two",
            "src",
            "dst",
        ],
        python_shell=False,
    )


def test_rsync_excludes_str():
    """
    Test for rsync files from src to dst with one exclude
    """
    mock = {
        "config.option": MagicMock(return_value=False),
        "cmd.run_all": MagicMock(),
    }
    with patch.dict(rsync.__salt__, mock):
        rsync.rsync("src", "dst", exclude="test/one")
    mock["cmd.run_all"].assert_called_once_with(
        ["rsync", "-avz", "--exclude", "test/one", "src", "dst"],
        python_shell=False,
    )
